# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
import os

import blinker
import urwid

from .group import Group
from .main import redraw_screen
from urwid import ScrollBar

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class CLIEditWidget(urwid.WidgetWrap):
    """Readline Edit widget with history and completion"""

    def __init__(self, prompt='', on_change=None, on_move=None, on_accept=None, on_cancel=None,
                 on_complete_next=None, on_complete_prev=None, completer=None,
                 history_file=None, history_size=1000, **kwargs):
        # Widgets
        self._editw = urwid.Edit(prompt, wrap='clip')
        self._candsw = CompletionCandidatesWidget(completer)
        self._groupw = Group(cls=urwid.Pile)
        self._groupw.add(name='candidates', widget=self._candsw, visible=False, options=10)
        self._groupw.add(name='cmdline', widget=self._editw, visible=True, options='pack')
        super().__init__(self._groupw)

        # Callbacks
        self._on_change = blinker.Signal()
        self._on_move = blinker.Signal()
        self._on_accept = blinker.Signal()
        self._on_cancel = blinker.Signal()
        self._on_complete_next = blinker.Signal()
        self._on_complete_prev = blinker.Signal()

        # Completion
        self._completer = completer
        self._on_complete_next.connect(self._cb_complete_next)
        self._on_complete_prev.connect(self._cb_complete_prev)
        self._on_change.connect(self._cb_change)
        self._on_move.connect(self._cb_move)
        self._user_input = None

        # History
        self._history = []
        self._history_size = history_size
        self._history_pos = -1
        self.history_file = history_file
        self._current_edit_text = ''

        # Internal callbacks
        self.on_cancel(lambda _: self.reset(), autoremove=False)

        # User callbacks
        if on_change is not None: self.on_change(on_change, autoremove=False)
        if on_move is not None: self.on_move(on_move, autoremove=False)
        if on_accept is not None: self.on_accept(on_accept, autoremove=False)
        if on_cancel is not None: self.on_cancel(on_cancel, autoremove=False)
        if on_complete_next is not None: self.on_complete_next(on_complete_next, autoremove=False)
        if on_complete_prev is not None: self.on_complete_prev(on_complete_prev, autoremove=False)

    def on_accept(self, callback, autoremove=True):
        self._on_accept.connect(callback, weak=autoremove)

    def on_cancel(self, callback, autoremove=True):
        self._on_cancel.connect(callback, weak=autoremove)

    def on_change(self, callback, autoremove=True):
        self._on_change.connect(callback, weak=autoremove)

    def on_move(self, callback, autoremove=True):
        self._on_move.connect(callback, weak=autoremove)

    def on_complete_next(self, callback, autoremove=True):
        self._on_complete_next.connect(callback, weak=autoremove)

    def on_complete_prev(self, callback, autoremove=True):
        self._on_complete_prev.connect(callback, weak=autoremove)

    def keypress(self, size, key):
        text_before = self._editw.edit_text
        curpos_before = self._editw.edit_pos

        callbacks = []
        cmd = self._editw._command_map[key]
        if cmd is urwid.CURSOR_UP:
            self._set_history_prev()
            key = None
        elif cmd is urwid.CURSOR_DOWN:
            self._set_history_next()
            key = None
        elif cmd is urwid.ACTIVATE:
            self.append_to_history()
            callbacks.append(self._on_accept)
            self._history_pos = -1
            key = None
        elif cmd is urwid.CANCEL:
            callbacks.append(self._on_cancel)
            self._history_pos = -1
            key = None
        elif cmd is urwid.COMPLETE_NEXT:
            callbacks.append(self._on_complete_next)
            key = None
        elif cmd is urwid.COMPLETE_PREV:
            callbacks.append(self._on_complete_prev)
            key = None
        elif key == 'space':
            key = super().keypress(size, ' ')
        else:
            key = super().keypress(size, key)

        if text_before != self._editw.edit_text:
            callbacks.append(self._on_change)
        elif curpos_before != self._editw.edit_pos:
            callbacks.append(self._on_move)

        for cb in callbacks:
            if cb is not None:
                cb.send(self)

        return key

    def reset(self):
        """Clear the command line and reset internal states"""
        self._editw.edit_text = ''
        self._groupw.hide('candidates')
        if self._completer is not None:
            self._completer.reset()
            self._candsw.reset()

    @property
    def edit_text(self):
        """Current editable text"""
        return self._editw.edit_text

    @edit_text.setter
    @redraw_screen
    def edit_text(self, cmd):
        self._editw.edit_text = str(cmd)

    @property
    def edit_pos(self):
        """Current cursor position in the editable text"""
        return self._editw.edit_pos

    @edit_pos.setter
    def edit_pos(self, pos):
        self._editw.edit_pos = int(pos)

    # Completion

    def _cb_complete_next(self, _):
        self._complete('next')

    def _cb_complete_prev(self, _):
        self._complete('prev')

    def _complete(self, direction):
        compl = self._completer
        if compl is not None:
            if direction == 'next':
                self._editw.edit_text, self._editw.edit_pos = compl.complete_next()
            elif direction == 'prev':
                self._editw.edit_text, self._editw.edit_pos = compl.complete_prev()
            else:
                raise ValueError('direction must be "next" or "prev"')
            log.debug('Completed: %r, %r', self._editw.edit_text, self._editw.edit_pos)
            self._candsw.update_candidates()

    def _cb_change(self, _):
        # Update candidate list whenever edit text is changed
        self._update_completion_candidates()

    def _cb_move(self, _):
        # Candidates are based on where the cursor is in the command line
        self._update_completion_candidates()

    def _update_completion_candidates(self):
        if self._completer is not None:
            # If the previously scheduled call isn't finished yet, abort it
            # because it's results will be out of date.
            old_task = getattr(self, '_completion_update_task', None)
            if old_task is not None and not old_task.done():
                old_task.cancel()

            def callback(task):
                try:
                    task.result()
                except asyncio.CancelledError:
                    pass
                else:
                    self._candsw.update_candidates()
                    self._maybe_hide_or_show_menu()

            coro = self._completer.update(self._editw.edit_text, self._editw.edit_pos)
            self._completion_update_task = asyncio.ensure_future(coro)
            self._completion_update_task.add_done_callback(callback)

    def _maybe_hide_or_show_menu(self):
        if self._completer is not None:
            candsw_visible = self._groupw.visible('candidates')
            any_matches = len(self._completer.categories) > 1  # First is current user input
            if any_matches and not candsw_visible:
                log.debug('Showing completion menu')
                self._groupw.show('candidates')
            elif not any_matches and candsw_visible:
                log.debug('Hiding completion menu')
                self._groupw.hide('candidates')

    # History

    # The most recent history entry is at index 0; -1 used to represent the current line
    # that is not yet in history and stored outside of self._history.

    def _set_history_prev(self):
        if self._history_pos == -1:
            self._current_edit_text = self._editw.edit_text
        if self._history_pos + 1 < len(self._history):
            self._history_pos += 1
            self._editw.edit_text = self._history[self._history_pos]
            self._editw.set_edit_pos(len(self._editw.edit_text))

    def _set_history_next(self):
        if self._history_pos > -1:
            self._history_pos -= 1
            if self._history_pos == -1:
                self._editw.edit_text = self._current_edit_text
            else:
                self._editw.edit_text = self._history[self._history_pos]
            self._editw.set_edit_pos(len(self._editw.edit_text))

    def _read_history(self):
        if self._history_file:
            self._history = list(reversed(_read_lines(self._history_file)))

    def append_to_history(self):
        """Add current `edit_text` to history"""
        # Don't add the same line twice in a row
        new_line = self._editw.edit_text
        if self._history and self._history[0] == new_line:
            return
        self._history.insert(0, new_line)
        if self._history_file is not None:
            _append_line(self._history_file, new_line)
        self._trim_history()

    def _trim_history(self):
        max_size = self._history_size
        if len(self._history) > max_size:
            # Trim history in memory
            self._history[:] = self._history[:max_size]

        if len(self._history) > 0 and self._history_file:
            # Trim history on disk
            flines = _read_lines(self._history_file)
            if len(flines) > max_size:
                # Trim more than necessary to reduce number of writes
                overtrim = max(0, min(int(self._history_size / 2), 10))
                flines = flines[overtrim:]
                _write_lines(self._history_file, flines)

    @property
    def history_file(self):
        return self._history_file

    @history_file.setter
    def history_file(self, path):
        if path is None:
            self._history_file = None
        else:
            if _mkdir(os.path.dirname(path)) and _test_access(path):
                self._history_file = os.path.abspath(os.path.expanduser(path))
                log.debug('Setting history_file=%r', self._history_file)
                self._read_history()

    @property
    def history_size(self):
        """Maximum number of lines kept in history"""
        return self._history_size

    @history_size.setter
    def history_size(self, size):
        self._history_size = int(size)
        self._trim_history()


class CompletionCandidatesWidget(urwid.WidgetWrap):
    def selectable(self):
        return False

    def __init__(self, completer):
        self._completer = completer
        self._listbox = urwid.ListBox(urwid.SimpleFocusListWalker([]))
        super().__init__(
            urwid.AttrMap(ScrollBar(
                urwid.AttrMap(self._listbox, 'completion')
            ), 'completion.scrollbar')
        )

    def reset(self):
        self._listbox.body[:] = ()

    def render(self, size, focus=False):
        # Always render as focused to highlight the focused candidate
        return super().render(size, focus=True)

    @redraw_screen
    def update_candidates(self):
        if not self._completer.categories:
            self.reset()
        else:
            cats = self._completer.categories

            def cat_widget_cols(cands):
                column_titles = []
                column_titles.append(cands.label)
                for cand in cands:
                    for title in cand.info:
                        if title not in column_titles:
                            column_titles.append(title)
                column_titles = tuple(urwid.Text(title) for title in column_titles)
                return urwid.Columns(column_titles)

            def cat_widget(cands):
                return urwid.AttrMap(urwid.Padding(cat_widget_cols(cands)), 'completion.category', 'default')

            def cand_widget(cand):
                if cand.in_parens:
                    cols = [urwid.Text('%s (%s)' % (cand, cand.in_parens))]
                else:
                    cols = [urwid.Text(cand)]
                for column_text in cand.info.values():
                    cols.append(urwid.Text(column_text))
                return urwid.AttrMap(urwid.Padding(urwid.Columns(cols)),
                                     'completion.item', 'completion.item.focused')

            widgets = []
            for cands in cats:
                if cands.label:
                    widgets.append(cat_widget(cands))
                for cand in cands:
                    widgets.append(cand_widget(cand))
            self._listbox.body[:] = widgets

            # Change focus position
            # Candidates are rendered as a flat list, which means we have to
            # calculate the focus_position, skipping category headers. Having
            # each category in a separate ListBox would be nicer, but nested
            # ListBoxes don't seem to work correctly: If the focused item is
            # moved out of view, the view doesn't scroll to follow it.
            cats = self._completer.categories
            if cats.current:
                pos = 0
                if cats.current_index > 0:
                    for cands in cats[:cats.current_index]:
                        if cands.label:
                            pos += 1
                        pos += len(cands)
                if cats.current.label:
                    pos += 1
                pos += cats.current.current_index
                self._listbox.focus_position = pos


def _mkdir(path):
    try:
        if path and not os.path.exists(path):
            os.makedirs(path)
    except OSError as e:
        log.error("Can't create directory {}: {}".format(path, e.strerror))
    else:
        return True


def _test_access(path):
    if path.startswith('/dev/null'):
        return True
    try:
        with open(path, 'a'): pass
    except OSError as e:
        log.error("Can't read/write history file {}: {}".format(path, e.strerror))
    else:
        return True


def _read_lines(filepath):
    if os.path.exists(filepath):
        try:
            with open(filepath, 'r') as f:
                return [line.strip() for line in f.readlines()]
        except OSError as e:
            log.error("Can't read history file {}: {}"
                      .format(filepath, e.strerror))
    return []


def _write_lines(filepath, lines):
    try:
        with open(filepath, 'w') as f:
            for line in lines:
                f.write(line.strip('\n') + '\n')
    except OSError as e:
        log.error("Can't write history file {}: {}".format(filepath, e.strerror))
    else:
        return True


def _append_line(filepath, line):
    try:
        with open(filepath, 'a') as f:
            f.write(line.strip() + '\n')
    except OSError as e:
        log.error("Can't append to history file {}: {}".format(filepath, e.strerror))
    else:
        return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...client import TorrentFilter
from .base import ItemWidgetBase, ListWidgetBase
from .torrent import TUICOLUMNS
from .utils import stringify_torrent_filter

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class TorrentItemWidget(ItemWidgetBase):
    palette_unfocused = 'torrentlist'
    palette_focused   = 'torrentlist.focused'
    columns_focus_map = {}
    for col in TUICOLUMNS.values():
        columns_focus_map.update(col.style.focus_map)

    @property
    def id(self):
        return self.data['id']

    @property
    def torrent_id(self):
        return self.data['id']


class TorrentListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = TorrentItemWidget
    keymap_context  = 'torrent'
    palette_name    = 'torrentlist'
    focusable_items = True

    def __init__(self, srvapi, keymap, tfilter=None, sort=None, columns=None, title=None):
        super().__init__(srvapi, keymap, columns=columns, sort=sort, title=title)
        self._tfilter = tfilter
        self._secondary_filter = None
        self._register_request()

    @property
    def id(self):
        """Hashable that is unique among all torrent lists"""
        return id(self)

    def _register_request(self):
        # Get keys needed for sort order, filters and columns
        keys = {'name'}
        if self._sort is not None:
            keys.update(self._sort.needed_keys)
        if hasattr(self._tfilter, 'needed_keys'):
            keys.update(self._tfilter.needed_keys)
        if self._secondary_filter is not None:
            keys.update(self._secondary_filter.needed_keys)
        for colname in self.columns:
            keys.update(self.tuicolumns[colname].needed_keys)

        # Register new request in request pool
        existing_keys = self._srvapi.treqpool.requested_keys(self.id)
        if keys != existing_keys:
            log.debug('Registering keys for %r: %s', self, keys)
            self._srvapi.treqpool.register(self.id,
                                           self._handle_torrents,
                                           keys=keys, tfilter=self._tfilter)
            self._srvapi.treqpool.poll()
        else:
            log.debug('No need to register a new request')
            self._invalidate()

    # # Enable this to measure rendering performance
    # def render(self, *args, **kwargs):
    #     import time
    #     start = time.time()
    #     canvas = super().render(*args, **kwargs)
    #     log.debug('Rendered torrent list in %.3fms', (time.time()-start)*1000)
    #     return canvas

    def _handle_torrents(self, torrents):
        # Auto-generate title from our filters if not set
        if self._title_name is None:
            self._title_name = stringify_torrent_filter(self._tfilter, torrents)
        self._data_dict = {t['id']:t for t in torrents}
        self._invalidate()

    def clear(self):
        for w in self._listbox.body:
            w.data.clearcache()
        super().clear()

    def refresh(self):
        self._srvapi.treqpool.poll()

    @property
    def sort(self):
        return self._sort

    @sort.setter
    def sort(self, sort):
        self._srvapi.treqpool.remove(self.id)
        ListWidgetBase.sort.fset(self, sort)
        self._register_request()

    @property
    def focused_torrent_id(self):
        """Torrent ID of the currently focused torrent or `None`"""
        focused_widget = self._listbox.focus
        if focused_widget is not None:
            return focused_widget.torrent_id

    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, torrent_filter):
        if torrent_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = TorrentFilter(torrent_filter)
        log.debug('Filtering %r torrents', self._secondary_filter)
        self._register_request()

    def _limit_items(self, torrent_widgets):
        f = self._secondary_filter
        if f is not None:
            for tw in torrent_widgets:
                if not f.match(tw.data):
                    yield tw



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ... import objects
from ..base import config as base
from . import _mixin as mixin
from ._table import print_table


class DumpCmd(base.DumpCmdbase):
    provides = {'cli'}

    def dump_rc(self, content, path, force=False):
        if path is None:
            print(content)
        else:
            return self.write_rc_file(content, path, force)


class RcCmd(base.RcCmdbase):
    provides = {'cli'}


class ResetCmd(base.ResetCmdbase):
    provides = {'cli'}


class SetCmd(base.SetCmdbase,
             mixin.only_supported_columns):
    provides = {'cli'}

    def make_setting_list(self, sort, columns):
        from ...views.setting import COLUMNS as SETTING_COLUMNS

        # Remove columns that aren't supported by CLI interface (e.g. 'marked')
        columns = self.only_supported_columns(columns, SETTING_COLUMNS)
        settings = sort.apply(objects.cfg.as_dict.values())
        print_table(settings, columns, SETTING_COLUMNS)


class RateLimitCmd(base.RateLimitCmdbase,
                   mixin.make_request, mixin.select_torrents):
    provides = {'cli'}

    async def _set_limits(self, TORRENT_FILTER, directions, limit, adjust=False, quiet=False):
        if len(TORRENT_FILTER) == 0 or TORRENT_FILTER == ['global']:
            await self._set_global_limits(directions, limit,
                                          adjust=adjust, quiet=quiet)
        else:
            await self._set_individual_limits(TORRENT_FILTER, directions, limit,
                                              adjust=adjust, quiet=quiet)

    async def _show_limits(self, TORRENT_FILTER, directions):
        if not TORRENT_FILTER or TORRENT_FILTER == ['global']:
            await self._show_global_limits(directions)
        else:
            await self._show_individual_limits(TORRENT_FILTER, directions)

    def _output(self, msg):
        print(msg)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import operator

from ..utils import Timedelta, Timestamp, const

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


def timestamp_or_timedelta(string, default_sign=1):
    """
    Try to parse `string` as Timestamp or Timedelta

    If `string` is parsed as Timedelta and the string does not indicate whether
    the Timedelta should be positive (in the future) or negative (in the past),
    multiply it by `default_sign`, which should be either 1 (default) or -1.
    """
    try:
        return Timestamp.from_string(string)
    except ValueError:
        delta = Timedelta.from_string(string)
        # Apply default_sign if no sign was given explicitly
        string = string.strip()
        if (not string.startswith('in') and
            not string.endswith('ago') and
            string[0] not in ('+', '-')):
            delta = delta.inverse if default_sign < 0 else delta
        return delta


_lgfilters = (operator.__lt__, operator.__le__, operator.__gt__, operator.__ge__)
def cmp_timestamp_or_timdelta(item_value, op, user_value):
    """Compare any combination of Timestamp and Timedelta objects"""
    # Some filters don't make any sense
    if (item_value in Timestamp.CONSTANTS and
        isinstance(user_value, Timedelta) and
        op in _lgfilters):
        return False
    elif item_value in Timedelta.CONSTANTS and op in _lgfilters:
        return False

    type_item_value = type(item_value)
    type_user_value = type(user_value)

    if type_user_value is Timestamp:
        if type_item_value is Timestamp:
            # result = op(item_value, user_value)
            # log.debug('1: %r %s %r = %r', item_value, op.__name__, user_value, result)
            # return result
            return op(item_value, user_value)
        elif type_item_value is Timedelta:
            # result = op(item_value, user_value.timedelta)
            # log.debug('2: %r %s %r (%r) = %r',
            #           item_value, op.__name__, user_value, user_value.timedelta, result)
            # return result
            return op(item_value, user_value.timedelta)
    elif type_user_value is Timedelta:
        # If the filter is, e.g., 'less than 1y ago', future dates would match
        # (because "[any future date] > -[seconds in 1y]") but we interpret the
        # filter as 'less than 1y ago up till now'.  Same thing for 'in less
        # than 1y' - past dates technically match, but they shouldn't.
        if not _either_past_or_future(item_value, user_value):
            return False
        elif type_item_value is Timedelta:
            # result = op(abs(item_value), abs(user_value))
            # log.debug('3: %r %s %r = %r', item_value, op.__name__, user_value, result)
            # return result
            return op(abs(item_value), abs(user_value))
        elif type_item_value is Timestamp:
            user_value_timestamp = user_value.timestamp
            if user_value < 0:
                if item_value in (Timestamp.NOW, Timestamp.SOON):
                    return False
                # result = op(user_value_timestamp, item_value)
                # log.debug('4: %r (%r) %s %r = %r',
                #           user_value, user_value_timestamp, op.__name__, item_value, result)
                # return result
                return op(user_value_timestamp, item_value)
            else:
                # result = op(item_value, user_value_timestamp)
                # log.debug('5: %r %s %r (%r) = %r',
                #           item_value, op.__name__, user_value, user_value_timestamp, result)
                # return result
                return op(item_value, user_value_timestamp)

    raise RuntimeError('cannot compare %r with %r' % (item_value, user_value))

def _either_past_or_future(item_value, user_value):
    """
    Return True if `item_value` and `user_value` are equal, both in the past or
    both in the future, False otherwise
    """
    type_item_value = type(item_value)
    type_user_value = type(user_value)

    if type_user_value is Timestamp:
        user_value_in_future = user_value.in_future
    elif type_user_value is Timedelta:
        user_value_in_future = user_value > 0

    if type_item_value is Timestamp:
        item_value_in_future = item_value.in_future
    elif type_item_value is Timedelta:
        item_value_in_future = item_value > 0

    return item_value_in_future == user_value_in_future


# TODO: Add docstring
def limit_rate_filter(limit, op, user_value):
    # `limit` may be a number or const.UNLIMITED.
    # `user_value` may be a number or a Bool (which is not a derivative of the
    # built-in `bool`).
    # If `user_value` is a Bool, True means 'limited' and False means
    # 'unlimited'.
    if isinstance(user_value, (int, float)):
        # This works because const.UNLIMITED behaves like `float('inf')`.
        return op(limit, user_value)
    else:
        if not user_value:
            # `user_value` is 'unlimited'/'off'/etc
            return op(limit, const.UNLIMITED)

        # `user_value` is 'limited'.
        elif op.__name__ in ('gt', 'ge'):
            # 'greater than limited' -> 'equal to unlimited'
            return limit == const.UNLIMITED
        elif op.__name__ in ('eq', 'ne'):
            # 'equal/unequal to limited' -> 'not equal/unequal to unlimited'
            return not op(limit, const.UNLIMITED)
        elif op.__name__ == 'lt':
            # 'lower than limited' -> No idea what to do here.
            return False
        elif op.__name__ == 'le':
            # Ignore the 'lower' in 'lower or equal';
            # 'equal to limited' -> ' unequal to unlimited'
            return limit != const.UNLIMITED
        else:
            # That should be all possible operators, but to avoid any
            # tracebacks, whatever is left gets regular treatment
            return op(limit, user_value)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import itertools
from collections import abc, defaultdict

from blinker import Signal

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class Settings(abc.Mapping):
    """Typed user configuration settings"""

    def __init__(self):
        self._defaults = {}
        self._values = {}
        self._constructors = {}
        self._getters = defaultdict(lambda: None)
        self._setters = defaultdict(lambda: None)
        self._descriptions = {}
        self._signals = defaultdict(lambda: Signal())
        self._global_signal = Signal()

    def add(self, name, constructor, default, description=None, getter=None, setter=None):
        """
        Add new setting

        name:        Identifier for this setting
        constructor: Callable that takes one argument and returns a new value
                     for this setting
        default:     Initial and default value
        description: What the setting does
        getter:      Callable with no arguments that returns the value
        setter:      Callable with one argument that sets the value
        """
        self._constructors[name] = constructor
        self._descriptions[name] = description
        value_ = self.validate(name, default)
        self._values[name] = value_
        self._defaults[name] = value_
        if getter:
            self._getters[name] = getter
        if setter:
            self._setters[name] = setter
        self._global_signal.send(self, name=name, value=value_)

    def reset(self, name):
        """Reset setting `name` to default/initial value"""
        self[name] = self._defaults[name]

    def default(self, name):
        """Return setting's default/initial value"""
        return self._defaults[name]

    def description(self, name):
        """Return setting's description"""
        return self._descriptions[name]

    def syntax(self, name):
        """Return setting's description"""
        return self._constructors[name].syntax

    def validate(self, name, value):
        """Pass `value` to `name`'s constructor and return the result"""
        if not isinstance(value, str) and isinstance(value, abc.Iterable):
            return self._constructors[name](*value)
        else:
            return self._constructors[name](value)

    def on_change(self, callback, name=None, autoremove=True):
        """
        Run `callback` every time a value changes

        If `name` is None, run `callback` after every change.

        The signature of `callback` must be: (settings, name, value)

        If `autoremove` is True, stop calling `callback` once it is garbage
        collected.
        """
        if name is None:
            self._global_signal.connect(callback, weak=autoremove)
        else:
            self._signals[name].connect(callback, weak=autoremove)

    @property
    def as_dict(self):
        # NOTE 1: The "id" key is important because filters (including
        #         SettingFilter) expect it.  This requirement can possibly
        #         removed but it probably means rewriting lots of filter tests.
        # NOTE 2: The "validate" lambda MUST store the value of `name`,
        #         otherwise `self.validate` always gets called with the for
        #         loop's last `name` value.
        return {name:{'id': name,
                      'value': self[name],
                      'default': self.default(name),
                      'description': self.description(name),
                      'syntax': self.syntax(name),
                      'validate': lambda v, n=name: self.validate(n, v)}
                for name,value in self.items()}

    def __getitem__(self, name):
        getter = self._getters[name]
        if getter is not None:
            return self.validate(name, getter())
        else:
            return self._values[name]

    def __setitem__(self, name, value):
        value_ = self.validate(name, value)
        setter = self._setters[name]
        if setter is not None:
            setter(value_)
        else:
            self._values[name] = value_
        self._global_signal.send(self, name=name, value=value_)
        self._signals[name].send(self, name=name, value=value_)

    def __contains__(self, name):
        return name in self._constructors

    def __iter__(self):
        return iter(self._constructors)

    def __len__(self):
        return len(self._constructors)


class RemoteSettings(abc.Mapping):
    """
    Thin wrapper around client.SettingsAPI that transparently adds/removes "srv." from
    keys and provides an interface similar to the Settings class
    """
    def __init__(self, remotecfg):
        self._cfg = remotecfg

    # Forward calls to SettingsAPI object

    def poll(self):
        """Get current values from server soon"""
        self._cfg.poll()

    async def update(self):
        """Get current values from server"""
        await self._cfg.update()

    async def set(self, name, value):
        """Change setting on the server"""
        if not name.startswith('srv.'):
            raise KeyError(name)
        await self._cfg.set(name[4:], value)

    def on_update(self, callback, autoremove=True):
        """Run `callback` after settings are updated"""
        return self._cfg.on_update(callback, autoremove=autoremove)

    def on_change(self, callback, name=None, autoremove=True):
        """
        Run `callback` every time a value is changed

        If `name` is None, run `callback` after every change.

        The signature of `callback` must be: (settings, name, value)

        If `autoremove` is True, stop calling `callback` once it is garbage
        collected.
        """
        self._cfg.on_set(callback, key=name, autoremove=autoremove)

    # Settings class protocol

    def reset(self, name):
        """Reset setting `name` to default/initial value"""
        if not name.startswith('srv.'):
            raise KeyError(name)
        raise NotImplementedError()

    def default(self, name):
        """Return setting's default/initial value"""
        if not name.startswith('srv.'):
            raise KeyError(name)
        return self._cfg.default(name[4:])

    def description(self, name):
        """Return setting's description"""
        if not name.startswith('srv.'):
            raise KeyError(name)
        return self._cfg.description(name[4:])

    def syntax(self, name):
        """Return setting's description"""
        if not name.startswith('srv.'):
            raise KeyError(name)
        return self._cfg.syntax(name[4:])

    def validate(self, name, value):
        """Pass `value` to `name`'s constructor and return the result"""
        if not name.startswith('srv.'):
            raise KeyError(name)
        return self._cfg.validate(name[4:], value)

    @property
    def as_dict(self):
        return {'srv.' + name: {**setting, 'id': 'srv.' + name}
                for name,setting in self._cfg.as_dict.items()}

    def __getitem__(self, name):
        if not name.startswith('srv.'):
            raise KeyError(name)
        return self._cfg[name[4:]]

    def __contains__(self, name):
        if not name.startswith('srv.'):
            return False
        return name[4:] in self._cfg

    def __iter__(self):
        return iter('srv.' + name for name in self._cfg)

    def __len__(self):
        return len(self._cfg)


class CombinedSettings(abc.Mapping):
    """
    Combination of Settings and RemoteSettings instances
    """
    def __init__(self, local, remote):
        self._local = local
        self._remote = remote

    def is_local(self, name):
        return not name.startswith('srv.')

    def is_remote(self, name):
        return name.startswith('srv.')

    async def update(self):
        """Get current remote settings from server"""
        await self._remote.update()

    async def set(self, name, value):
        if name in self._local:
            self._local[name] = value
        elif name in self._remote:
            await self._remote.set(name, value)
        else:
            raise KeyError(name)

    def _find(self, name):
        if name in self._local:
            return self._local
        elif name in self._remote:
            return self._remote
        else:
            raise KeyError(name)

    def reset(self, name):
        """Reset setting `name` to default/initial value"""
        return self._find(name).reset(name)

    def default(self, name):
        """Return setting's default/initial value"""
        return self._find(name).default(name)

    def description(self, name):
        """Return setting's description"""
        return self._find(name).description(name)

    def syntax(self, name):
        """Return setting's description"""
        return self._find(name).syntax(name)

    def validate(self, name, value):
        """Pass `value` to `name`'s constructor and return the result"""
        return self._find(name).validate(name, value)

    @property
    def as_dict(self):
        return {**self._local.as_dict, **self._remote.as_dict}

    def __getitem__(self, name):
        return self._find(name)[name]

    def __setitem__(self, name, value):
        self._find(name)[name] = value

    def __contains__(self, name):
        return name in self._local or name in self._remote

    def __iter__(self):
        return iter(itertools.chain(self._local, self._remote))

    def __len__(self):
        return len(self._local) + len(self._remote)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from collections import abc, defaultdict

import urwid

from ..utils.string import strwidth
from .main import redraw_screen

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class TabID(int):
    def __repr__(self):
        return '<TabID %d>' % self


def _find_unused_id(existing_ids):
    """Find lowest unused ID in `existing_ids`"""
    if not existing_ids:
        return TabID(0)
    else:
        for id_candidate in range(0, max(existing_ids) + 2):
            if id_candidate not in existing_ids:
                return TabID(id_candidate)


class TabBar(urwid.GridFlow):
    def __init__(self, spacing=1, default_width=20):
        return super().__init__([], default_width, spacing, 0, 'left')

    def __getitem__(self, pos):
        return self.contents[pos][0]

    def __setitem__(self, pos, widget):
        self.contents[pos] = self._make_title(widget)

    def __delitem__(self, pos):
        del self.contents[pos]

    def insert(self, pos, widget):
        self.contents.insert(pos, self._make_title(widget))

    def append(self, widget):
        self.contents.append(self._make_title(widget))

    def _make_title(self, widget):
        if hasattr(widget.base_widget, 'text'):
            opts = ('given', strwidth(widget.base_widget.text))
        else:
            opts = ()
        return (widget, self.options(*opts))

    @property
    def focus(self):
        return self.contents.focus

    @focus.setter
    def focus(self, pos):
        self.contents.focus = pos

    def __iter__(self):
        for w in self.contents:
            yield w[0]


class Tabs(urwid.Widget):
    """Organize multiple widgets in tabs"""

    _sizing = frozenset([urwid.FLOW, urwid.BOX])
    _max_focus_history_size = 100

    def __init__(self, *contents, tabbar=None):
        """
        Create new Tabs widget

        contents: Iterable of dictionaries or iterables that match the arguments
                  of the `insert` method
        tabbar: TabBar instance that is used to display tab titles or any object
                with a 'base_widget' attribute (e.g. AttrMap) that returns a
                TabBar object
        """
        if tabbar is None:
            self._tabbar = TabBar()
        elif not isinstance(tabbar, urwid.Widget):
            raise ValueError('tabbar must be TabBar instance, not {}: {!r}'
                             .format(type(tabbar).__name__, tabbar))
        else:
            self._tabbar = tabbar

        self._ids = []
        self._focus_history = []
        self._info = defaultdict(lambda: {})
        self._contents = urwid.MonitoredFocusList()
        self._contents.set_focus_changed_callback(self._focus_changed_callback)
        for content in contents:
            if not isinstance(content, abc.Mapping):
                content = dict(zip(('title', 'widget', 'position', 'focus'),
                                   content))
            self.insert(**content)

    def render(self, size, focus=False):
        if len(size) < 2:
            cols, rows = (size[0], None)
        else:
            cols, rows = size

        if len(self._contents) < 1:
            # No contents - return empty canvas
            return urwid.SolidCanvas(' ', cols, rows)

        if rows is not None:
            size_content = (cols, rows - self._tabbar.rows((cols,)))
        else:
            size_content = (cols,)

        combinelist = []
        position = self._contents.focus

        # Always render tab titles as focused to highlight the focused tab
        canvas = self._tabbar.render((cols,), focus=True)
        combinelist.append((canvas, position, True))

        # Render and add content of currently selected tab
        current_widget = self._contents[position]
        if current_widget is None:
            canvas = urwid.SolidCanvas(' ', *size_content)
        else:
            canvas = current_widget.render(size_content, focus)
        combinelist.append((canvas, position, focus))
        return urwid.CanvasCombine(combinelist)

    def get_index(self, position=None):
        """
        Return tab index at `position` or None if there are no tabs

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        if position is None:
            return self.focus_position
        elif isinstance(position, TabID):
            if position in self._ids:
                return self._ids.index(position)
            else:
                raise IndexError('No tab with ID: {}'.format(position))
        else:
            i = self.focus_position if position is None else position
            if i is not None:
                tab_count = len(self._contents)
                if i < 0:
                    # Negative index means right-bound index (-1 is the rightmost tab)
                    if i >= -tab_count:
                        return self.get_index(tab_count + i)
                elif i < tab_count:
                    # Positive index means left-bound index (1 is the leftmost tab)
                    return i
                raise IndexError('No tab at position: {}'.format(position))

    def get_id(self, position=None):
        """
        Return unique TabID of tab at `position` or None if there are no tabs

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        return self._ids[i] if i is not None else None

    def load(self, title, widget=None, position=None, focus=True):
        """
        Set content at `position`, in focused tab or in new tab

        If `position` is not None, it is forwarded to `set_title`/`set_content`
        together with `title`/`widget`.

        If no tabs exist, a new tab is created with `title` and `widget`.  If
        `widget` is None, a blank widget is used.

        Otherwise, the focused tab's title and content is replaced with
        `set_title` and `set_content`.

        Set `focus` to False to load content in background.

        Return TabID of newly created tab
        """
        if position is not None:
            # Overload content in specified tab
            self.set_content(widget, position=position)
            self.set_title(title, position=position)
            if focus:
                if isinstance(position, TabID):
                    self.focus_id = position
                else:
                    self.focus_position = position
            return self.get_id(position)
        elif self.focus_position is None:
            # No tabs exist - create new tab
            return self.insert(title, widget, focus=focus)
        else:
            # Overload content in focused tab
            self.set_content(widget)
            self.set_title(title)
            return self.get_id()

    @redraw_screen
    def insert(self, title, widget=None, position=-1, focus=True):
        """
        Insert new tab

        title: Any flow or fixed widget to use as the tab's title
        widget: Widget to show when this tab is selected or None
        position: Where to insert the new tab; int for list-like index or
                  'right'/'left' to insert next to focused tab
        focus: True to focus the new tab, False otherwise

        Return TabID of inserted tab
        """
        curpos = self.focus_position
        if position == 'right':
            newpos = (curpos + 1) if curpos is not None else 0
        elif position == 'left':
            newpos = max(curpos, 0) if curpos is not None else 0
        elif isinstance(position, int):
            if position < 0:
                newpos = position + len(self._contents) + 1
            else:
                newpos = position
        else:
            raise ValueError('Invalid position: {!r}'.format(position))

        # Insert new tab ID
        this_id = _find_unused_id(self._ids)
        self._ids.insert(newpos, this_id)

        # Insert title
        self._tabbar.base_widget.insert(newpos, title)

        # Insert content
        self._contents.insert(newpos, widget)
        if focus:
            self.focus_position = newpos
        return this_id

    @redraw_screen
    def move(self, position=None, destination='right', wrap=False):
        """
        Move tab at `position` to `where`

        position: Index (int), ID (TabID) or None (focused tab)
        destination: "left", "right" or new index (int)
        wrap: Whether to move the right-most tab to the first index when it is
              moved to the right and the left-most tab to the last index when it
              is moved to the left

        Raises IndexError if tab can't be found.
        """
        focused_tab_id = self.focus_id
        max_index = len(self._ids) - 1
        old_index = self.get_index(position)
        if destination == 'left':
            new_index = old_index - 1
            if new_index < 0:
                new_index = max_index if wrap else 0
        elif destination == 'right':
            new_index = old_index + 1
            if new_index > max_index:
                new_index = 0 if wrap else max_index
        elif isinstance(destination, int):
            new_index = destination
        else:
            raise ValueError('Invalid destination: %r' % (destination,))

        # Temporarily disable focus change callback
        self._contents.set_focus_changed_callback(lambda _: None)

        for lst in (self._ids, self._contents, self._tabbar.base_widget):
            item = lst[old_index]
            del lst[old_index]
            if new_index == -1:
                # list.insert() can only insert before
                lst.append(item)
            elif new_index < -1:
                # First negative index is -1
                lst.insert(new_index + 1, item)
            else:
                lst.insert(new_index, item)

        # Restore focus
        self.focus_id = focused_tab_id
        self._contents.set_focus_changed_callback(self._focus_changed_callback)

    @redraw_screen
    def remove(self, position=None):
        """
        Remove tab `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        index = self.get_index(position)
        tabid = self.get_id(position)
        if tabid is None:
            raise RuntimeError('Tabs is empty')
        del self._ids[index]
        del self._contents[index]
        if tabid in self._info:
            del self._info[tabid]
        del self._tabbar.base_widget[index]
        fh = self._focus_history
        while tabid in fh:
            fh.remove(tabid)

    def clear(self):
        """Remove all tabs"""
        while len(self._ids):
            self.remove(0)  # Remove tab at index 0

    def get_title(self, position=None):
        """
        Return tab title widget at `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        if i is not None:
            return self._tabbar.base_widget[i]

    def set_title(self, title, position=None):
        """
        Change the title widget of a tab

        title: New title widget; should be a Text object optionally wrapped by
               AttrMap
        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        if i is not None:
            self._tabbar.base_widget[i] = title
        else:
            raise RuntimeError('Tabs is empty')

    def get_content(self, position=None):
        """
        Return tab content widget at `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        if i is not None:
            return self._contents[i]

    def set_content(self, widget=None, position=None):
        """
        Set content of tab at `position` to `widget`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        if i is not None:
            self._contents[i] = widget
        else:
            raise RuntimeError('Tabs is empty')

    def get_info(self, position=None):
        """
        Return information about tab at `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        tabid = self.get_id(position)
        if tabid is not None:
            return self._info[tabid]

    def set_info(self, position=None, **kwargs):
        """
        Set information about tab at `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Information is taken as arbitrary keyword arguments and is stored as an ordinary
        dictionary.

        Raises IndexError if tab can't be found.
        """
        tabid = self.get_id(position)
        if tabid is not None:
            self._info[tabid].update(kwargs)
        else:
            raise RuntimeError('Tabs is empty')

    def _focus_changed_callback(self, pos):
        tab_id = self.get_id(pos)
        self._focus_history.append(tab_id)
        while len(self._focus_history) > self._max_focus_history_size:
            self._focus_history.pop(0)

    @property
    def focus(self):
        """Content widget of currently focused tab or None if no tabs exist"""
        position = self._contents.focus
        if position is not None:
            return self._contents[position]
        return None

    @property
    def prev_focus(self):
        """Content widget of previously focused tab or None"""
        fh = self._focus_history
        if len(fh) >= 2:
            # fh[-1] is the currently focused tab ID
            prev_id = fh[-2]
            try:
                return self.get_content(prev_id)
            except IndexError:
                pass

    @property
    def focus_position(self):
        """Index (starting from 0) of currently focused tab or None if no tabs exist"""
        return self._contents.focus

    @focus_position.setter
    def focus_position(self, position):
        if 0 <= position < len(self._contents):
            self._tabbar.base_widget.focus = position
            self._contents.focus = position
        else:
            raise IndexError('No tab at position: {!r}'.format(position))

    @property
    def prev_focus_position(self):
        """Index of previously focused tab or None"""
        fh = self._focus_history
        if len(fh) >= 2:
            # fh[-1] is the currently focused tab ID
            prev_id = fh[-2]
            try:
                return self.get_index(prev_id)
            except IndexError:
                pass

    @property
    def focus_id(self):
        """TabID of currently focused tab or None if no tabs exist"""
        i = self.focus_position
        if i is not None:
            return self._ids[i]

    @focus_id.setter
    @redraw_screen
    def focus_id(self, tabid):
        i = self.get_index(tabid)
        if 0 <= i < len(self._contents):
            self._tabbar.base_widget.focus = i
            self._contents.focus = i
        else:
            raise IndexError('No tab with ID: {}'.format(tabid))

    @property
    def prev_focus_id(self):
        """TabID of previously focused tab or None"""
        fh = self._focus_history
        if len(fh) >= 2:
            return fh[-2]

    @property
    def ids(self):
        """Yields all tab IDs"""
        for w in self._ids:
            yield w

    @property
    def contents(self):
        """Yields all content widgets"""
        for w in self._contents:
            yield w

    @property
    def titles(self):
        """Yields all tab title widgets"""
        yield from self._tabbar.base_widget

    def __len__(self):
        return len(self._contents)

    def __iter__(self):
        return iter(self._contents)

    def selectable(self):
        return True

    def keypress(self, size, key):
        focus_widget = self.focus

        if focus_widget is not None:
            if len(size) > 1:
                cols, rows = size
                size = (cols, rows - self._tabbar.rows((cols,)))
            else:
                size = (size[0],)

            if focus_widget.selectable():
                key = focus_widget.keypress(size, key)

        if key is not None:
            focus_pos = self.focus_position
            cmd = self._command_map[key]
            if cmd == urwid.CURSOR_LEFT and focus_pos > 0:
                self.focus_position -= 1
                key = None
            elif cmd == urwid.CURSOR_RIGHT and focus_pos < len(self._contents) - 1:
                self.focus_position += 1
                key = None
        return key



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from .. import rdns
from .base import SorterBase, SortSpec


class _SortSpec(SortSpec):
    def __init__(self, *args, description='', **kwargs):
        description = 'Sort peers by %s' % description
        super().__init__(*args, description=description, **kwargs)


def _get_hostname_or_ip(torrent):
    ip = torrent['ip']
    hostname = rdns.gethostbyaddr_from_cache(ip)
    return ip if hostname is None else hostname


class PeerSorter(SorterBase):
    DEFAULT_SORT = 'torrent'
    SORTSPECS = {
        'torrent'     : _SortSpec(lambda t: t['tname'].casefold(),
                                  description='torrent name'),
        '%downloaded' : _SortSpec(lambda t: t['%downloaded'],
                                  aliases=('%dn',),
                                  description='download progress'),
        'rate-up'     : _SortSpec(lambda t: t['rate-up'],
                                  aliases=('rup',),
                                  description='upload rate (from our perspective)'),
        'rate-down'   : _SortSpec(lambda t: t['rate-down'],
                                  aliases=('rdn',),
                                  description='download rate (from our perspective)'),
        'rate-est'    : _SortSpec(lambda t: t['rate-est'],
                                  aliases=('re',),
                                  description='estimated overall download rate'),
        'rate'        : _SortSpec(lambda t: t['rate-up'] + t['rate-down'],
                                  aliases=('r',),
                                  description='combined download and upload rate'),
        'eta'         : _SortSpec(lambda t: t['eta'],
                                  description='estimated remaining download time'),
        'client'      : _SortSpec(lambda t: t['client'].casefold(),
                                  aliases=('cl',),
                                  description='client name'),
        'host'        : _SortSpec(_get_hostname_or_ip,
                                  description='IP address or hostname'),
        'port'        : _SortSpec(lambda t: t['port'],
                                  description='port number'),
    }



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering TorrentFiles by various values"""

from ..ttypes import TorrentFile
from .base import BoolFilterSpec, CmpFilterSpec, Filter, FilterChain, FilterSpecDict


class _BoolFilterSpec(BoolFilterSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, needed_keys=('files',), **kwargs)

class _CmpFilterSpec(CmpFilterSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, needed_keys=('files',), **kwargs)

class _SingleFilter(Filter):
    DEFAULT_FILTER = 'name'

    BOOLEAN_FILTERS = FilterSpecDict({
        'all'      : _BoolFilterSpec(None,
                                     aliases=('*',),
                                     description='All files'),
        'wanted'   : _BoolFilterSpec(lambda f: f['is-wanted'],
                                     description='Wanted files'),
        'complete' : _BoolFilterSpec(lambda f: f['%downloaded'] >= 100,
                                     aliases=('cmp',),
                                     description='Fully downloaded files'),
    })

    COMPARATIVE_FILTERS = FilterSpecDict({
        'name'        : _CmpFilterSpec(value_getter=lambda f: f['name'],
                                       value_type=TorrentFile.TYPES['name'],
                                       aliases=('n',),
                                       description='Match VALUE against file name'),
        'path'        : _CmpFilterSpec(value_getter=lambda f: f['path-absolute'],
                                       value_type=TorrentFile.TYPES['path-absolute'],
                                       aliases=('dir',),
                                       description='Match VALUE against file path'),
        'size'        : _CmpFilterSpec(value_getter=lambda f: f['size-total'],
                                       value_type=TorrentFile.TYPES['size-total'],
                                       aliases=('sz',),
                                       description='Match VALUE against file size'),
        'downloaded'  : _CmpFilterSpec(value_getter=lambda f: f['size-downloaded'],
                                       value_type=TorrentFile.TYPES['size-downloaded'],
                                       as_bool=lambda f: f['%downloaded'] >= 100,
                                       aliases=('dn',),
                                       description='Match VALUE against downloaded bytes'),
        '%downloaded' : _CmpFilterSpec(value_getter=lambda f: f['%downloaded'],
                                       value_type=TorrentFile.TYPES['%downloaded'],
                                       as_bool=lambda f: f['%downloaded'] >= 100,
                                       aliases=('%dn',),
                                       description='Match VALUE against percentage of downloaded bytes'),
        'priority'    : _CmpFilterSpec(value_getter=lambda f: f['priority'],
                                       value_type=TorrentFile.TYPES['priority'],
                                       as_bool=lambda f: f['priority'] != 0,  # Any non-normal priority
                                       aliases=('prio',),
                                       description='Match VALUE against download priority (off, low, normal, high)'),
    })


class FileFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = _SingleFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
import os

from ... import objects
from ...completion import candidates
from ...utils.cliparser import Arg
from .. import CmdError, CommandMeta
from . import _mixin as mixin
from ._common import (make_COLUMNS_doc, make_SCRIPTING_doc, make_SORT_ORDERS_doc,
                      make_X_FILTER_spec)

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class AddTorrentsCmdbase(metaclass=CommandMeta):
    name = 'add'
    aliases = ('download','get')
    provides = set()
    category = 'torrent'
    description = 'Download torrents'
    usage = ('add [<OPTIONS>] <TORRENT> <TORRENT> <TORRENT> ...',)
    examples = ('add 72d7a3179da3de7a76b98f3782c31843e3f818ee',
                'add --stopped http://example.org/something.torrent',
                'add --labels linux,iso https://archlinux.org/releng/releases/2022.04.05/torrent/')
    argspecs = (
        {'names': ('TORRENT',), 'nargs': '+',
         'description': 'Link or path to torrent file, magnet link or info hash'},

        {'names': ('--stopped','-s'), 'action': 'store_true',
         'description': 'Do not start downloading the added torrent(s)'},

        {'names': ('--path','-p'),
         'description': ('Custom download directory for added torrent(s) '
                         'relative to "srv.path.complete" setting')},

        {'names': ('--labels','-l'),
         'description': 'Comma-separated list of labels'},
    )

    async def run(self, TORRENT, stopped, path, labels):
        success = True
        force_torrentlist_update = False
        if labels:
            labels = labels.split(',')
        for source in TORRENT:
            source_abs_path = self.make_path_absolute(source)
            response = await self.make_request(objects.srvapi.torrent.add(source_abs_path,
                                                                          stopped=stopped,
                                                                          path=path,
                                                                          labels=labels))
            success = success and response.success
            force_torrentlist_update = force_torrentlist_update or success

        # Update torrentlist AFTER all 'add' requests
        if force_torrentlist_update and hasattr(self, 'polling_frenzy'):
            self.polling_frenzy()

        if not success:
            raise CmdError()

    @staticmethod
    def make_path_absolute(path):
        abspath = os.path.abspath(os.path.expanduser(path))
        if os.path.exists(abspath):
            return abspath
        else:
            return path

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option == '--path':
            return candidates.fs_path(args.curarg.before_cursor,
                                      base=objects.cfg['srv.path.complete'],
                                      directories_only=True)
        if option == '--labels':
            curlbl = args.curarg.split(',')[-1]
            return candidates.labels(curlbl)


class TorrentDetailsCmdbase(mixin.get_single_torrent, metaclass=CommandMeta):
    name = 'details'
    aliases = ('info',)
    provides = set()
    category = 'torrent'
    description = 'Display detailed torrent information'
    usage = ('details',
             'details <TORRENT FILTER>')
    examples = ('details id=71',)
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
    )

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True,
                                           prefer_focused=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            torrent = await self.get_single_torrent(tfilter, keys=('id', 'name'))
            if not torrent:
                raise CmdError()
            else:
                log.debug('Showing details of torrent %r: %r', tfilter, torrent)
                if asyncio.iscoroutinefunction(self.display_details):
                    await self.display_details(torrent['id'])
                else:
                    self.display_details(torrent['id'])

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        if args.curarg_index == 1:
            return candidates.torrent_filter(args.curarg)


class ListTorrentsCmdbase(mixin.get_torrent_sorter, mixin.get_torrent_columns,
                          metaclass=CommandMeta):
    name = 'list'
    aliases = ('ls',)
    provides = set()
    category = 'torrent'
    description = 'List torrents'
    usage = ('list [<OPTIONS>]',
             'list [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('ls active',
                'ls !active',
                'ls seeds<10',
                'ls active&tracker~example.org',
                'ls active|idle&tracker~example')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=False, nargs='*'),

        {'names': ('--sort', '-s'),
         'default_description': "current value of 'sort.torrents' setting",
         'description': ('Comma-separated list of sort orders '
                         "(see SORT ORDERS section)")},

        {'names': ('--columns', '-c'),
         'default_description': "current value of 'columns.torrents' setting",
         'description': ('Comma-separated list of column names '
                         "(see COLUMNS section)")},
    )

    from ...client.sorters import TorrentSorter
    from ...views.torrent import COLUMNS
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.torrents'),
        'SORT ORDERS': make_SORT_ORDERS_doc(TorrentSorter, '--sort', 'sort.torrents'),
        'SCRIPTING': make_SCRIPTING_doc(name),
    }

    async def run(self, TORRENT_FILTER, sort, columns):
        sort = objects.localcfg['sort.torrents'] if sort is None else sort
        columns = objects.localcfg['columns.torrents'] if columns is None else columns
        try:
            columns = self.get_torrent_columns(columns)
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=True,
                                           discover_torrent=False)
            sort = self.get_torrent_sorter(sort)
        except ValueError as e:
            raise CmdError(e)
        else:
            log.debug('Listing %s torrents sorted by %s', tfilter, sort)
            if asyncio.iscoroutinefunction(self.make_torrent_list):
                await self.make_torrent_list(tfilter, sort, columns)
            else:
                self.make_torrent_list(tfilter, sort, columns)

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.torrent_filter(args.curarg)

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option == '--sort':
            return candidates.sort_orders('TorrentSorter')
        elif option == '--columns':
            return candidates.column_names('torrents')


class TorrentMagnetURICmdbase(metaclass=CommandMeta):
    name = 'magnet'
    aliases = ('uri',)
    provides = set()
    category = 'torrent'
    description = 'Display torrent(s) magnet URI'
    usage = ('magnet',
             'magnet <TORRENT FILTER>')
    examples = ('magnet name~ubuntu',)
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
    )

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True,
                                           prefer_focused=False)
        except ValueError as e:
            raise CmdError(e)
        else:
            try:
                uris = await objects.srvapi.torrent.get_magnet_uris(tfilter)
            except objects.srvapi.ClientError as e:
                raise CmdError(e)
            else:
                self.display_uris(uris)

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        if args.curarg_index == 1:
            return candidates.torrent_filter(args.curarg)


class MoveTorrentsCmdbase(metaclass=CommandMeta):
    name = 'move'
    aliases = ('mv',)
    provides = set()
    category = 'torrent'
    description = "Change torrents' location"
    usage = ('move <PATH>',
             'move <TORRENT FILTER> <PATH>')
    examples = ('move ./new/path',
                'move size>50G /path/to/lots/of/storage')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),

        {'names': ('PATH',),
         'description': ('Move the specified torrent(s) to this directory.  If PATH is relative '
                         '(i.e. does not start with "/"), it is relative to the value of the '
                         'setting "srv.path.complete".  That means "." is the download path.')},
    )

    async def run(self, TORRENT_FILTER, PATH):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(objects.srvapi.torrent.move(tfilter, PATH),
                                               polling_frenzy=True)
            if not response.success:
                raise CmdError()


class RemoveTorrentsCmdbase(metaclass=CommandMeta):
    name = 'remove'
    aliases = ('rm', 'delete')
    provides = set()
    category = 'torrent'
    description = 'Remove torrents'
    usage = ('remove [<OPTIONS>]',
             'remove [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('remove',
                r'remove "stupid torrent" silly\ torrent and_this_torrent',
                'remove -d "unwanted torrent"')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),

        {'names': ('--delete-files','-d'), 'action': 'store_true',
         'description': 'Delete any downloaded files'},

        {'names': ('--force','-f'), 'action': 'store_true',
         'description': ('Ignore remove.max-hits setting: Remove all '
                         'matching torrents instead of asking for confirmation '
                         'if the number of matches exceeds remove.max-hits')},
    )

    async def run(self, TORRENT_FILTER, delete_files, force):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            async def do_remove(tfilter=tfilter, delete_files=delete_files):
                response = await self.make_request(
                    objects.srvapi.torrent.remove(tfilter, delete=delete_files),
                    polling_frenzy=True)
                if not response.success:
                    raise CmdError()

            async def do_keep(tfilter=tfilter):
                self.error(('Keeping %s torrents: Too many hits ' % tfilter) +
                           '(use --force or increase remove.max-hits setting)')

            response = await objects.srvapi.torrent.torrents(tfilter, keys=('id',))
            hits = len(response.torrents)
            success = hits > 0
            if force or objects.cfg['remove.max-hits'] < 0 or hits < objects.cfg['remove.max-hits']:
                return await do_remove()
            else:
                await self.show_list_of_hits(tfilter)
                if hits > 0:
                    question = 'Are you sure you want to remove %d torrent%s' % (
                        hits, '' if hits == 1 else 's')
                    if delete_files:
                        question += ' and their files'
                    question += '?'
                    success = await self.ask_yes_no(question, yes=do_remove, no=do_keep,
                                                    after=self.remove_list_of_hits)
                if not success:
                    raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.torrent_filter(args.curarg)


class RenameCmdbase(metaclass=CommandMeta):
    name = 'rename'
    aliases = ('rn',)
    provides = set()
    category = 'torrent'
    description = 'Rename a torrent or one of its files or directories'
    usage = ('rename <NEW>',
             'rename <TORRENT> <NEW>')
    examples = ('rename "A Better Name"',
                'rename id=123 Foo',
                'rename id=123/some/file new_file_name')
    argspecs = (
        {'names': ('TORRENT',), 'nargs': '?',
         'description': ('Torrent filter expression, optionally followed by a "/" and '
                         'the relative path to a file or directory in the torrent'),
         'default_description': 'Focused torrent, file or directory in the TUI'},

        {'names': ('NEW',),
         'description': ('New name of the torrent, file or directory specified by TORRENT '
                         '(must not contain "/" or be "." or "..")')},

        {'names': ('--unique', '-u'), 'action': 'store_true',
         'description': ('Ensure the torrent filter expression in TORRENT matches exactly '
                         'one torrent; if not given, all matching files in all matching '
                         'torrents are renamed'),
         'default_description': 'Enabled automatically when renaming torrents'},
    )

    async def run(self, TORRENT, NEW, unique):
        if not TORRENT:
            # Autodetect path
            path = self.get_relative_path_from_focused(unique=unique)
            if path:
                # path is "<TORRENT
                # IDENTIFIER>/relative/path/to/file/in/torrent" where <TORRENT
                # IDENTIFIER> is either the torrent's name or "id=<ID>" if
                # `unique` is True.
                TORRENT = path

        # Split torrent filter from relative path in torrent
        if TORRENT and '/' in TORRENT:
            FILTER, PATH = TORRENT.split('/', maxsplit=1)
            renaming_torrent = False
        else:
            FILTER, PATH = TORRENT, None
            renaming_torrent = True

        try:
            tfilter = self.select_torrents(FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(
                objects.srvapi.torrent.torrents(tfilter, keys=('id',)),
                quiet=True)
            if not response.success:
                raise CmdError()
            elif (unique or renaming_torrent) and len(response.torrents) > 1:
                # When renaming a torrent or --unique is given, tfilter must
                # match exactly one torrent.  If it matches zero torrents,
                # make_request() below with produce the appropriate error
                # message.
                raise CmdError('%s matches more than one torrent' % tfilter)
            else:
                success = True
                for torrent in response.torrents:
                    tid = torrent['id']
                    response = await self.make_request(
                        objects.srvapi.torrent.rename(tid, path=PATH, new_name=NEW),
                        polling_frenzy=True)
                    if not response.success:
                        success = False
                if not success:
                    raise CmdError()

    @classmethod
    async def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        # We don't care about options
        args = args.posargs()
        if args.curarg_index == 1:
            # Complete file and directory names from torrent(s)
            return await candidates.torrent_path(args.curarg, only='any')
        elif args.curarg_index == 2:
            first_arg_stripped = args[1].strip('/')
            if '/' in first_arg_stripped:
                # First argument contains a path separator, so destination is
                # the new file name.  The second argument can't contain a path
                # separator if it's a file or directory.
                if '/' not in args.curarg:
                    # To make it more convenient to adjust file/directory names,
                    # destination candidates are existing files or directories
                    # from the path specified in the first argument.  Files if
                    # the first argument points to a file, directories if the
                    # first argument points to a directory.
                    source = Arg(first_arg_stripped, curpos=len(first_arg_stripped))
                    log.debug('Using destination candidates from: %r', source)
                    return await candidates.torrent_path(source, only='auto')
            else:
                # Destination is the new torrent name
                log.debug('Using torrent names as destination candidates')
                return await candidates.torrent_filter(args.curarg, filter_names=False)


# Argument definitions that are shared between commands
ARGSPEC_TOGGLE = {
    'names': ('--toggle','-t'), 'action': 'store_true',
    'description': ('Start TORRENT if stopped and vice versa')
}

class StartTorrentsCmdbase(metaclass=CommandMeta):
    name = 'start'
    aliases = ()
    provides = set()
    category = 'torrent'
    description = 'Start downloading torrents'
    usage = ('start [<OPTIONS>]',
             'start [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('start',
                "start 'night of the living dead' Metropolis",
                'start ubuntu --force')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),

        {'names': ('--force','-f'), 'action': 'store_true',
         'description': 'Ignore download queue'},

        ARGSPEC_TOGGLE,
    )

    async def run(self, TORRENT_FILTER, toggle, force):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            if toggle:
                response = await self.make_request(
                    objects.srvapi.torrent.toggle_stopped(tfilter, force=force),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    objects.srvapi.torrent.start(tfilter, force=force),
                    polling_frenzy=True)
            if not response.success:
                raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.torrent_filter(args.curarg)


class StopTorrentsCmdbase(metaclass=CommandMeta):
    name = 'stop'
    aliases = ('pause',)
    provides = set()
    category = 'torrent'
    description = 'Stop downloading torrents'
    usage = ('stop [<OPTIONS>]',
             'stop [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('stop',
                'stop "night of the living dead" idle',
                'stop --toggle ubuntu')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),
        ARGSPEC_TOGGLE,
    )

    async def run(self, TORRENT_FILTER, toggle):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            if toggle:
                response = await self.make_request(
                    objects.srvapi.torrent.toggle_stopped(tfilter),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    objects.srvapi.torrent.stop(tfilter),
                    polling_frenzy=True)
            if not response.success:
                raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.torrent_filter(args.curarg)


class VerifyTorrentsCmdbase(metaclass=CommandMeta):
    name = 'verify'
    aliases = ('check',)
    provides = set()
    category = 'torrent'
    description = 'Verify downloaded torrent data'
    usage = ('verify [<OPTIONS>]',
             'verify [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('verify',
                'verify debian')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),
    )

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(objects.srvapi.torrent.verify(tfilter),
                                               polling_frenzy=True)
            if not response.success:
                raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.torrent_filter(args.curarg)


class LabelCmd(metaclass=CommandMeta):
    name = 'label'
    provides = set()
    category = 'torrent'
    description = 'Manipulate torrent labels'
    usage = ('label [<OPTIONS>] <TORRENT FILTER> <TORRENT FILTER>... <[LABEL][,LABEL...]>',)
    examples = ('label iso,linux id=34',
                'label -r iso,linux id=34',
                'label -c id=34')
    argspecs = (
        {'names': ('LABELS',),
         'description': ('Comma-separated list of labels to add/remove'),
         'nargs': '?', 'default': ''},
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),
        {'names': ('--clear','-c'), 'action': 'store_true',
         'description': 'Clear all labels'},
        {'names': ('--remove','-r'), 'action': 'store_true',
         'description': 'Remove labels rather than adding them'},
        {'names': ('--set','-s'), 'action': 'store_true',
         'description': 'Set labels to exactly the LABELS argument',
         'dest': '_set'},
        {'names': ('--quiet','-q'), 'action': 'store_true',
         'description': 'Do not show new label(s)'},
    )

    async def run(self, LABELS, TORRENT_FILTER, _set, remove, clear, quiet):
        if not (_set ^ remove ^ clear) and (_set or remove or clear):
            raise CmdError('At most one of --set/s, --remove/r, --clear,-c can be present.')
            return

        if clear:
            def handler(tfilter, labels):
                return objects.srvapi.torrent.labels_clear(tfilter)
        elif _set:
            handler = objects.srvapi.torrent.labels_set
        elif remove:
            handler = objects.srvapi.torrent.labels_remove
        else:
            handler = objects.srvapi.torrent.labels_add

        labels = LABELS.split(',')

        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)

        response = await self.make_request(handler(tfilter, labels),
                                           polling_frenzy=True, quiet=quiet)
        if not response.success:
            raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        curlbl = args.curarg.split(',')[-1]
        return candidates.labels(curlbl)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import os
import re
from collections.abc import Iterable
from functools import partial
from itertools import chain
from typing import Pattern

_INFINITY = float('inf')


def _resolve_alias(value, aliases):
    # Only hashable values can be aliases
    try:
        hash(value)
    except Exception:
        return value
    else:
        # Return original value if it doesn't have an alias
        return aliases.get(value, value)


def _pretty_float(n):
    n_abs = abs(n)
    if n_abs >= _INFINITY:
        return ('-' if n < 0 else '') + '∞'
    elif n_abs == 0:
        return '0'
    n_abs_r2 = round(n_abs, 2)
    if n_abs_r2 == int(n_abs):
        return '%.0f' % n
    elif n_abs_r2 < 10:
        return ('%.2f' % n).rstrip('0').rstrip('.')
    elif round(n_abs, 1) < 100:
        return ('%.1f' % n).rstrip('0').rstrip('.')
    else:
        return '%.0f' % n


class _PartialConstructor(partial):
    def __init__(self, cls, **kwargs):
        repr = cls.__name__ + '('
        if kwargs:
            repr += ', '.join('%s=%r' % (k,v) for k,v in kwargs.items())
        self.__repr = repr + ')'
        self.__name__ = cls.__name__

    @property
    def syntax(self):
        return self.func._get_syntax(**self.keywords)

    @property
    def typename(self):
        return self.func.typename

    def __repr__(self):
        return self.__repr


class UsertypeMixin():
    @classmethod
    def partial(cls, **kwargs):
        return _PartialConstructor(cls, **kwargs)

    def __init__(self, *value, **kwargs):
        self._config = {**self.defaults, **kwargs}

    def copy(self, *value, **kwargs):
        new_kwargs = {**self._config, **kwargs}
        new_posargs = value if len(value) > 0 else self
        if not isinstance(new_posargs, Iterable):
            new_posargs = (new_posargs,)
        return type(self)(*new_posargs, **new_kwargs)

    @property
    def syntax(self):
        return self._get_syntax(**self._config)


# https://stackoverflow.com/a/5192374
class classproperty():
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


def multitype(*constructors):
    class MultitypeMeta(type):
        def __new__(mcls, name, bases, clsattrs):
            cls = type.__new__(mcls, name, bases, clsattrs)
            cls.__name__ = 'Or'.join(subcls.__name__ for subcls in cls._subclses)
            mcls._subclses = (cls,) + cls._subclses
            return cls

        @classmethod
        def __instancecheck__(mcls, inst):
            for subtype in mcls._subclses:
                if type(inst) is subtype:
                    return True
            return False

        @classmethod
        def __subclasscheck__(mcls, cls):
            return cls in mcls._subclses[1:]


    class Multitype(UsertypeMixin, metaclass=MultitypeMeta):
        # Get constructors (e.g. Int.partial(...)) for making new values
        _constructors = tuple(c if isinstance(c, _PartialConstructor) else c.partial()
                              for c in constructors)

        # Get classes for instance checking (see MultitypeMeta)
        _subclses = tuple(c.func for c in _constructors)

        def __new__(cls, *value, **kwargs):
            self = cls._get_instance(*value, **kwargs)
            # Overload syntax string
            self._get_syntax = lambda cls=cls, **_: cls.syntax
            self.typename = cls.typename
            return self

        @classmethod
        def _get_instance(cls, *value, **kwargs):
            errors = []
            for const in cls._constructors:
                try:
                    return const(*value)
                except ValueError as e:
                    errors.append(str(e))
                except TypeError:
                    errors.append('Not a %s' % const.typename)
            raise ValueError('; '.join(errors))

        @classproperty
        def typename(cls):
            return ' or '.join((s.typename for s in cls._subclses if s.typename))

        @classproperty
        def syntax(cls):
            return ' or '.join((c.syntax for c in cls._constructors))

        @classmethod
        def _get_syntax(cls, **_):
            return cls.syntax

    return Multitype


class String(str, UsertypeMixin):
    """
    String

    Options:
      minlen: Minimum length of the string
      maxlen: Maximum length of the string
      regex:  Regular expression values must match or None
    """
    typename = 'string'
    defaults = {'minlen': 0,
                'maxlen': _INFINITY,
                'regex': None}

    def __new__(cls, value, *, minlen=defaults['minlen'], maxlen=defaults['maxlen'],
                regex=defaults['regex']):
        # Convert
        self = super().__new__(cls, value)

        # Validate
        self_len = len(self)
        if maxlen is not None and self_len > maxlen:
            raise ValueError('Too long (maximum length is %s)' % maxlen)
        if minlen is not None and self_len < minlen:
            raise ValueError('Too short (minimum length is %s)' % minlen)
        if regex is not None:
            if not self._ensure_regex(regex).search(self):
                raise ValueError('Invalid value')
        return self

    @classmethod
    def _get_syntax(cls, minlen=defaults['minlen'], maxlen=defaults['maxlen'],
                    regex=defaults['regex']):
        if ((minlen == 1 or minlen <= 0) and
            (maxlen == 1 or maxlen >= _INFINITY)):
            chrstr = 'character'
        else:
            chrstr = 'characters'

        specs = []
        if regex is not None:
            specs.append('pattern=%s' % (cls._ensure_regex(regex).pattern,))
        if minlen > 0 and maxlen < _INFINITY:
            if minlen == maxlen:
                specs.append('%s %s' % (minlen, chrstr))
            else:
                specs.append('%s-%s %s' % (minlen, maxlen, chrstr))
        elif minlen > 0:
            specs.append('at least %s %s' % (minlen, chrstr))
        elif maxlen < _INFINITY:
            specs.append('at most %s %s' % (maxlen, chrstr))

        if specs:
            return 'string (%s)' % (', '.join(specs),)
        else:
            return 'string'

    @staticmethod
    def _ensure_regex(regex):
        if isinstance(regex, Pattern):
            return regex
        else:
            try:
                return re.compile(regex)
            except re.error:
                raise RuntimeError('Invalid regex: %r' % (regex,))


class Bool(str, UsertypeMixin):
    """
    Boolean

    Options:
      true: Sequence of truthy values
      false: Sequence of falsy values
    """
    typename = 'boolean'
    defaults = {'true'  : ('enabled', 'yes', 'on', 'true', '1'),
                'false' : ('disabled', 'no', 'off', 'false', '0')}

    def __new__(cls, value, *, true=defaults['true'], false=defaults['false']):
        if isinstance(value, str):
            _value = value.casefold()
        else:
            _value = value

        # Validate
        if _value in true:
            is_true = True
        elif _value in false:
            is_true = False
        elif isinstance(_value, bool):
            is_true = _value
            if is_true:
                value = true[0]
            else:
                value = false[0]
        else:
            raise ValueError('Not a %s' % cls.typename)

        self = super().__new__(cls, value)
        self._is_true = is_true
        return self

    @staticmethod
    def _get_syntax(true=defaults['true'], false=defaults['false']):
        pairs = []
        for pair in zip(true, false):
            pair = tuple(str(val) for val in pair)
            if pair not in pairs:
                pairs.append(pair)
        return '%s' % '|'.join('/'.join((t,f)) for t,f in pairs)

    def __bool__(self):
        return self._is_true

    def __eq__(self, other):
        if isinstance(other, bool):
            return other == self._is_true
        elif isinstance(other, type(self)):
            return other._is_true == self._is_true
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._is_true)

    @property
    def truths(self):
        return self._config['true']

    @property
    def falsities(self):
        return self._config['false']


class Path(str, UsertypeMixin):
    """
    File system path

    Options:
      mustexist: Whether the path must exist on the local file system
      path:      Base path for relative path values
    """
    typename = 'path'
    defaults = {'mustexist': False, 'base': ''}

    def __new__(cls, value, *, base=defaults['base'], mustexist=defaults['mustexist']):
        # Convert
        self = super().__new__(cls, os.path.expanduser(os.path.normpath(value)))

        # Validate
        if mustexist and not os.path.exists(self):
            raise ValueError('No such file or directory')

        # Paths that start with './' or '../' are considered absolute
        self._is_relative = (not os.path.isabs(value) and
                             not str(value).startswith('.' + os.sep) and
                             not str(value).startswith('..' + os.sep))
        self._base = os.path.expanduser(base)
        return self

    @staticmethod
    def _get_syntax(base=defaults['base'], mustexist=defaults['mustexist']):
        if base:
            return 'file system path relative to %s' % Path(base).prettified
        else:
            return 'file system path'

    @property
    def prettified(self):
        home = os.path.expanduser('~')
        if self.startswith(home):
            return '~' + self[len(home):]
        else:
            return str(self)

    @property
    def base_path(self):
        return self._base

    @property
    def full_path(self):
        if self._is_relative:
            return type(self)(os.path.join(self._base, self))
        return self


class Tuple(tuple, UsertypeMixin):
    """
    Immutable list

    Options:
      sep:     Separator between list items in string representation
      options: Iterable of valid values; any other values raise ValueError
      aliases: <alias> -> <value> mapping: any occurrence of <alias> is replaced
               with <value>
      dedup:   Whether to remove duplicate items
    """
    typename = 'list'
    defaults = {'sep'     : ', ',
                'options' : None,
                'aliases' : {},
                'dedup'   : False}

    def __new__(cls, *value, sep=defaults['sep'], options=defaults['options'],
                aliases=defaults['options'], dedup=defaults['dedup']):
        def normalize(val):
            if isinstance(val, str):
                for item in val.split(sep.strip()):
                    yield item.strip()
            else:
                yield val

        # Convert
        value = (chain.from_iterable((normalize(item) for item in value)))
        if aliases:
            value = (_resolve_alias(item, aliases) for item in value)
        if dedup:
            _seen = set()
            value = (item for item in value if not (item in _seen or _seen.add(item)))
        self = super().__new__(cls, value)

        if options is not None:
            # A single '*' replaces it with all available options
            if len(self) == 1 and self[0] == '*':
                self = super().__new__(cls, options)

            # Validate
            invalid_items = tuple(str(item) for item in self if item not in options)
            if invalid_items:
                raise ValueError('Invalid option%s: %s' % (
                    's' if len(invalid_items) != 1 else '',
                    sep.join(invalid_items)))

        return self

    @staticmethod
    def _get_syntax(sep=defaults['sep'], **_):
        sep = sep.strip()
        return '<OPTION>%s<OPTION>%s...' % (sep, sep)

    def __str__(self):
        return self._config['sep'].join(str(item) for item in self)

    @property
    def sep(self):
        return self._config['sep']

    @property
    def options(self):
        return self._config['options']

    @property
    def aliases(self):
        return self._config['aliases']

    @property
    def aliases_inverse(self):
        """Mapping of values to lists of aliases"""
        options = {}
        for alias,option in self._config['aliases'].items():
            aliases = options.get(option, [])
            aliases.append(alias)
            options[option] = aliases
        return options


class Option(str, UsertypeMixin):
    """
    Single string that can only be one of a given set of strings

    Options:
      options: Iterable of valid values; any other values raise ValueError
      aliases: <alias> -> <value> mapping; any occurrence of <alias> is replaced
               with <value>
    """
    typename = 'option'
    defaults = {'options': (),
                'aliases': {}}

    def __new__(cls, value, *, options=defaults['options'], aliases=defaults['aliases']):
        value = str(value)
        value = _resolve_alias(value, aliases)

        if not any(value == option for option in options):
            if len(options) == 0:
                raise RuntimeError('No options provided')
            elif len(options) == 1:
                raise ValueError('Not %s' % options[0])
            else:
                raise ValueError('Not one of: %s' % ', '.join((str(o) for o in options)))

        self = super().__new__(cls, value)
        return self

    @staticmethod
    def _get_syntax(options=defaults['options'], aliases=defaults['aliases']):
        return '|'.join(str(opt) for opt in options)

    @property
    def options(self):
        return self._config['options']

    @property
    def aliases(self):
        return self._config['aliases']

    @property
    def aliases_inverse(self):
        """Mapping of options to lists of aliases"""
        options = {}
        for alias,option in self._config['aliases'].items():
            aliases = options.get(option, [])
            aliases.append(alias)
            options[option] = aliases
        return options


class _NumberBase(UsertypeMixin):
    _prefixes_binary = (('Ti', 1024**4), ('Gi', 1024**3), ('Mi', 1024**2), ('Ki', 1024))
    _prefixes_metric = (('T', 1000**4), ('G', 1000**3), ('M', 1000**2), ('k', 1000))
    _prefixes_dct = {prefix.lower():size
                     for prefix,size in chain.from_iterable(zip(_prefixes_binary,
                                                                _prefixes_metric))}
    _regex = re.compile(r'^\s*([-+]?(?:\d+\.\d+|\d+|\.\d+|inf)) ?(' +
                        r'|'.join(p[0] for p in chain.from_iterable(
                            zip(_prefixes_binary, _prefixes_metric))) +
                        r'|)([^\s0-9]*?)\s*$',
                        flags=re.IGNORECASE)

    converters = {
        'B': {'b': lambda value: value * 8},  # bytes to bits
        'b': {'B': lambda value: value / 8},  # bits to bytes
    }
    typename = 'number'
    defaults = {'unit'       : None,
                'convert_to' : None,
                'prefix'     : None,
                'hide_unit'  : None,
                'min'        : None,
                'max'        : None,
                'autolimit'  : None}

    def __new__(cls, value, *, unit=defaults['unit'],
                convert_to=defaults['convert_to'], prefix=defaults['prefix'],
                hide_unit=defaults['hide_unit'], min=defaults['min'],
                max=defaults['max'], autolimit=defaults['autolimit']):
        if isinstance(value, cls):
            # Use value's arguments as defaults
            defaults = value._args
            unit = unit if unit is not None else defaults['unit']
            prefix = prefix if prefix is not None else defaults['prefix']
            hide_unit = hide_unit if hide_unit is not None else defaults['hide_unit']
            value = float(value)

        # Fill in hardcoded defaults
        prefix = 'metric' if prefix is None else prefix
        hide_unit = False if hide_unit is None else hide_unit

        # Parse strings
        if isinstance(value, str):
            string = str(value)
            match = cls._regex.match(string)
            if match is None:
                raise ValueError('Not a %s' % cls.typename)
            else:
                value = float(match.group(1))
                prfx = match.group(2)
                unit = match.group(3) or unit
                if prfx:
                    value *= cls._prefixes_dct[prfx.lower()]

                prfx_len = len(prfx)
                if prfx_len == 2:
                    prefix = 'binary'
                elif prfx_len == 1:
                    prefix = 'metric'

        # Scale number to different unit
        if convert_to is not None and unit != convert_to:
            if unit is None:
                # num has no unit - assume num is already in target unit
                unit = convert_to
            else:
                converters = cls.converters
                if unit in converters and convert_to in converters[unit]:
                    converter = converters[unit][convert_to]
                    value = converter(value)
                    unit = convert_to
                else:
                    raise ValueError('Cannot convert %s to %s' % (unit, convert_to))

        if issubclass(cls, int):
            try:
                value = round(value)
            except OverflowError:
                raise ValueError('Not a %s' % cls.typename)

        if min is not None and value < min:
            if autolimit:
                value = min
            else:
                raise ValueError('Too small (minimum is %s)' % min)
        elif max is not None and value > max:
            if autolimit:
                value = max
            else:
                raise ValueError('Too big (maximum is %s)' % max)

        try:
            self = super().__new__(cls, value)
        except TypeError:
            raise ValueError('Not a %s' % cls.typename)

        if hide_unit:
            self._str = lambda: self.without_unit
        else:
            self._str = lambda: self.with_unit

        if prefix == 'binary':
            self._prefixes = self._prefixes_binary
        elif prefix == 'metric':
            self._prefixes = self._prefixes_metric
        elif prefix == 'none':
            self._prefixes = ()
        else:
            raise ValueError("prefix must be 'binary' or 'metric'")

        # Remember arguments so we can copy them if this instance is passed to the same class
        self._args = {'unit': unit, 'prefix': prefix, 'hide_unit': hide_unit,
                      'min': min, 'max': max, 'autolimit': autolimit}
        return self

    @classmethod
    def _get_syntax(cls, **_):
        prefixes = (p[0] for p in chain(cls._prefixes_binary, cls._prefixes_metric))
        return '<NUMBER>[%s]' % '|'.join(prefixes)

    def __str__(self):
        return self._str()

    @property
    def with_unit(self):
        """String representation including unit"""
        s = self.without_unit
        unit = self._args['unit']
        if unit is not None:
            s += unit
        return s

    @property
    def without_unit(self):
        """String representation excluding unit"""
        absolute = abs(self)
        if self == 0:
            # This should increase efficiency since 0 is a common value
            return '0'
        elif absolute >= _INFINITY:
            return _pretty_float(self)
        else:
            for prefix,size in self._prefixes:
                if absolute >= size:
                    # Converting to float/int before doing the math is faster
                    # because we overload math operators.
                    return _pretty_float(float(self) / size) + prefix
            return _pretty_float(self)

    @property
    def unit(self):
        return self._args['unit']

    @property
    def hide_unit(self):
        return self._args['hide_unit']

    @property
    def prefix(self):
        return self._args['prefix']

    def _do_math(self, funcname, other=None, **kwargs):
        # If self and other have a unit specified, convert other if possible, or
        # raise an exception.
        if other is not None and isinstance(other, _NumberBase) and self.unit != other.unit:
            other = other.copy(other, convert_to=self.unit)

        # Get the new value as int or float
        if self >= _INFINITY:
            # No need to do anything with infinity because `int` has no infinity
            # value implemented.
            result = _INFINITY
        else:
            result = getattr(float(self), funcname)(other, **kwargs)

        if result is NotImplemented:
            return NotImplemented
        elif not hasattr(result, 'is_integer') or result.is_integer():
            result_cls = Int
        else:
            result_cls = Float

        # Create new instance with copied properties
        return result_cls(result, **self._args)

    def __add__(self, other):          return self._do_math('__add__', other)
    def __sub__(self, other):          return self._do_math('__sub__', other)
    def __mul__(self, other):          return self._do_math('__mul__', other)
    def __div__(self, other):          return self._do_math('__div__', other)
    def __truediv__(self, other):      return self._do_math('__truediv__', other)
    def __floordiv__(self, other):     return self._do_math('__floordiv__', other)
    def __mod__(self, other):          return self._do_math('__mod__', other)
    def __divmod__(self, other):       return self._do_math('__divmod__', other)
    def __pow__(self, other):          return self._do_math('__pow__', other)
    def __floor__(self):               return self._do_math('__floor__')
    def __ceil__(self):                return self._do_math('__ceil__')
    def __round__(self, ndigits=None): return self._do_math('__round__', ndigits)

class Float(_NumberBase, float):
    """
    Floating point number

    TODO: ...
    """

class Int(_NumberBase, int):
    """
    Integer number

    TODO: ...
    """

class Percent(Float):
    """Float with 'unit' argument defaulting to '%'"""

    def __new__(cls, num, *, unit='%', **kwargs):
        return super().__new__(cls, num, unit=unit, **kwargs)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..base import FreeSpaceAPIBase

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class FreeSpaceAPI(FreeSpaceAPIBase):
    async def get_free_space(self, path):
        """Return free space in directory `path` in bytes"""
        response = await self._rpc.free_space(path=path)
        log.debug('Free space in %r: %r', path, response)
        if path == response['path']:
            return response['size-bytes']
        else:
            raise RuntimeError('Expected path %r, got %r' % (path, response['path']))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...client import TrackerFilter
from .base import ItemWidgetBase, ListWidgetBase
from .tracker import TUICOLUMNS
from .utils import stringify_torrent_filter

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class TrackerItemWidget(ItemWidgetBase):
    palette_unfocused = 'trackerlist'
    palette_focused   = 'trackerlist.focused'
    columns_focus_map = {}
    for col in TUICOLUMNS.values():
        columns_focus_map.update(col.style.focus_map)

    @property
    def id(self):
        return self.data['id']

    @property
    def torrent_id(self):
        return self.data['tid']


class TrackerListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = TrackerItemWidget
    keymap_context  = 'tracker'
    palette_name    = 'trackerlist'
    focusable_items = True

    def __init__(self, srvapi, keymap, torfilter, trkfilter, columns=None, sort=None, title=None):
        super().__init__(srvapi, keymap, columns=columns, sort=sort, title=title)
        self._torfilter = torfilter
        self._trkfilter = trkfilter
        self._secondary_filter = None

        # Create tracker filter generator
        if trkfilter is not None:
            def filter_trackers(trackers):
                yield from trkfilter.apply(trackers)
        else:
            def filter_trackers(trackers):
                yield from trackers
        self._maybe_filter_trackers = filter_trackers

        self._poller = self._srvapi.create_poller(
            self._srvapi.torrent.torrents, torfilter, keys=('trackers', 'name', 'id')
        )
        self._poller.on_response(self._handle_trackers)

    def _handle_trackers(self, response):
        if response is None or not response.torrents:
            self.clear()
        else:
            # Auto-generate title from our filters if not set
            if self._title_name is None:
                self._title_name = stringify_torrent_filter(self._torfilter, response.torrents)
                if self._trkfilter:
                    self._title_name += ' %s' % self._trkfilter

            # Create list items our base widget can handle
            def trackers_combined(torrents):
                for t in torrents:
                    yield from self._maybe_filter_trackers(t['trackers'])
            self._data_dict = {trk['id']:trk for trk in trackers_combined(response.torrents)}
        self._invalidate()

    def refresh(self):
        self._poller.poll()

    @property
    def sort(self):
        return self._sort

    @sort.setter
    def sort(self, sort):
        ListWidgetBase.sort.fset(self, sort)
        self._poller.poll()

    @property
    def focused_torrent_id(self):
        """Torrent ID of the currently focused tracker or `None`"""
        focused_widget = self._listbox.focus
        if focused_widget is not None:
            return focused_widget.torrent_id


    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, tracker_filter):
        if tracker_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = TrackerFilter(tracker_filter)
        self._invalidate()

    def _limit_items(self, tracker_widgets):
        # Combine primary and secondary tracker filters
        trkfilter = self._trkfilter
        strkfilter = self._secondary_filter
        if trkfilter is None:
            trkfilter = strkfilter
        elif strkfilter is not None:
            trkfilter = trkfilter & strkfilter

        if trkfilter is not None:
            for tw in tracker_widgets:
                if not trkfilter.match(tw.data):
                    log.debug('%r does not match %r', trkfilter, tw.data['domain'])
                    yield tw
                else:
                    log.debug('%r does match %r', trkfilter, tw.data['domain'])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes for TUI commands"""

import asyncio
import functools
import os

from ... import client, objects
from .. import CmdError, utils
from ._common import make_tab_title_widget

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


def _deep_getattr(obj, *attrs):
    for attr in attrs:
        obj = getattr(obj, attr, None)
        if obj is None:
            break
    return obj


class make_request():
    async def make_request(self, request_coro, polling_frenzy=False, quiet=False):
        """
        Awaits `request_coro` and optionally logs messages to `process`

        If `polling_frenzy` evaluates to True and the request succeeded, the
        `polling_frenzy` method is called.

        Return the response object returned by `request_coro`.
        """
        response = await request_coro
        utils.log_msgs(self, response, quiet)
        if response.success and polling_frenzy:
            self.polling_frenzy()
        return response


class ask_yes_no():
    ANSWERS = {'y': True, 'n': False,
               'Y': True, 'N': False}

    async def ask_yes_no(self, question, yes=None, no=None, after=None):
        """
        Ask user a yes/no question

        The `yes` and `no` arguments are callbacks (or None) that are called
        depending on the user's answer. `after` is called after the user
        answered and after `yes` or `no` has been called.

        Callbacks may be normal functions, coroutine functions or
        coroutines. They don't get any arguments and their return value is
        ignored.
        """
        from ...tui import tuiobjects

        def run_func_or_coro(func_or_coro):
            if asyncio.iscoroutinefunction(func_or_coro):
                asyncio.ensure_future(func_or_coro())
            elif asyncio.iscoroutine(func_or_coro):
                asyncio.ensure_future(func_or_coro)
            elif func_or_coro is not None:
                func_or_coro()

        class YesNoEditWidget(tuiobjects.urwid.Edit):
            def keypress(slf, size, key):
                answer = self.ANSWERS.get(key, None)
                if answer is not None:
                    tuiobjects.widgets.remove('yesnoprompt')
                    tuiobjects.widgets.focus_name = focus_name
                    if answer:
                        run_func_or_coro(yes)
                    else:
                        run_func_or_coro(no)
                    run_func_or_coro(after)
                return None

        # Remember focused widget
        focus_name = tuiobjects.widgets.focus_name

        widget = tuiobjects.urwid.AttrMap(YesNoEditWidget(question + ' [y|n] '), 'prompt')
        pos = tuiobjects.widgets.get_position('main') + 1
        tuiobjects.widgets.add(widget=widget, name='yesnoprompt', removable=True,
                               options='pack', position=pos)


class select_torrents():
    def select_torrents(self, FILTER, allow_no_filter=True, discover_torrent=True, prefer_focused=False):
        """
        Get TorrentFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, the result of `discover_torrent_ids` is
        returned if `discover_torrent` evaluates to True.  `prefer_focused` is
        forwarded to `discover_torrent_ids`.  If `discover_torrent` evaluates to
        False, None is returned if `allow_no_filter` evaluates to True,
        otherwise a ValueError is raised.
        """
        if FILTER:
            return client.TorrentFilter(FILTER)
        else:
            if discover_torrent:
                tids = self.discover_torrent_ids(prefer_focused=prefer_focused)
                if tids is not None:
                    return self.ids2tfilter(tids)
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent specified')

    def discover_torrent_ids(self, prefer_focused=False):
        """
        Auto-detect which torrents are currently selected by the user

        Torrents are selected by marking them or by focusing them.  If
        `prefer_focused` evaluates to True, the focused torrent ID is returned,
        if possible, before looking for marks.  Otherwise, the focused torrent
        ID is only returned if there are no marked torrents.

        If any torrents are found, return a tuple of torrent IDs, otherwise
        None.
        """
        def get_marked():
            tids = self.get_marked_torrent_ids()
            if tids:
                log.debug('Found marked torrents: %r', tids)
                return tids

        def get_focused():
            tid = self.get_focused_torrent_id()
            if tid:
                log.debug('Found focused torrent: %r', tid)
                return (tid,)

        if prefer_focused:
            return get_focused() or get_marked()
        else:
            return get_marked() or get_focused()

    def get_marked_torrent_ids(self):
        """
        Return IDs of marked items in the current or previous tab

        This relies on the widget having a `marked` attribute.
        """
        widget = self._get_current_or_previous_tab()
        if hasattr(widget, 'marked'):
            tids = tuple(twidget.torrent_id for twidget in widget.marked)
            if tids:
                return set(tids)

    @classmethod
    def get_focused_torrent_id(cls):
        """
        Return torrent ID of focused item in the current or previous tab

        This relies on the current widget having a `focused_torrent_id`
        attribute.
        """
        widget = cls._get_current_or_previous_tab()
        if hasattr(widget, 'focused_torrent_id'):
            tid = widget.focused_torrent_id
            if tid is not None:
                return tid


    @classmethod
    def get_focused_item_data(cls):
        """
        Return 'torrent', 'file' or 'directory' depending on what is currently
        focused
        """
        from ...tui.tuiobjects import tabs
        return _deep_getattr(tabs, 'focus', 'focused_widget', 'data')

    @classmethod
    def get_focused_item_type(cls):
        """
        Return 'torrent', 'file' or 'directory' depending on what is currently
        focused
        """
        from ...tui.tuiobjects import tabs
        focused_widget = _deep_getattr(tabs, 'focus', 'focused_widget')
        focused_data = _deep_getattr(focused_widget, 'data')

        if isinstance(focused_data, client.Torrent):
            return 'torrent'
        from ...views.file import TorrentFileDirectory
        if isinstance(focused_data, TorrentFileDirectory):
            return 'directory'
        if isinstance(focused_data, client.TorrentFile):
            return 'file'
        from ...tui.views import SettingItemWidget
        if isinstance(focused_widget, SettingItemWidget):
            return 'setting'

    @classmethod
    def _get_current_or_previous_tab(cls):
        """Return currently focused tab content if not empty, the previous one or None"""
        from ...tui.tuiobjects import tabs
        widget = tabs.focus
        if widget is not None:
            return widget
        else:
            return tabs.prev_focus

    @staticmethod
    def ids2tfilter(ids):
        """Turn an iterable of ids into a TorrentFilter instance"""
        if all(isinstance(x, int) for x in ids):
            return client.TorrentFilter('|'.join(('id=%d' % id for id in ids)))
        else:
            raise RuntimeError('Not a list of IDs: %r' % (ids,))


class select_files():
    def get_relative_path_from_focused(self, unique=False):
        """
        When viewing a file list, return '<TORRENT NAME>/<RELATIVE PATH>' of the
        focused file or directory, or None when viewing anything else

        If `unique` evaluates to True, '<TORRENT NAME>' is replaced with
        'id=<TORRENT ID>' to make this path unique to this torrent even if there
        are multiple torrents with the same name.
        """
        from ...tui.tuiobjects import tabs
        focused_widget = tabs.focus
        if hasattr(focused_widget, 'focused_file_ids'):
            data = focused_widget.focused_widget.data
            path = data['path-relative']
            if unique:
                # Remove torrent name from path
                first_slash = path.find('/')
                if first_slash > -1:
                    # Multi-file torrent
                    return 'id=%d/%s' % (data['tid'], path[first_slash:].strip('/'))
                else:
                    # Single-file torrent
                    return 'id=%d' % (data['tid'],)
            else:
                return path

    def select_files(self, FILTER, allow_no_filter=True, discover_file=True):
        """
        Get FileFilter instance, focused/marked file IDs or None

        If `FILTER` evaluates to True, it is passed to FileFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False and `discover_file` evaluates to True,
        a tuple of (<torrent ID>, <file ID>) tuples is returned.  If
        `discover_file` evaluates to False, None is returned if
        `allow_no_filter` evaluates to True, else a ValueError is raised.

        Files are discovered by getting the `focused_file_ids` of the focused
        tab's widget.
        """
        if FILTER:
            return client.FileFilter(FILTER)
        else:
            if discover_file:
                fids = self._find_file_ids()
                if fids:
                    log.debug('Found file IDs: %r', fids)
                    return fids
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent file specified')

    def _find_file_ids(self):
        from ...tui.tuiobjects import tabs
        focused_widget = tabs.focus
        # Get marked file IDs
        if hasattr(focused_widget, 'marked'):
            fids = tuple(fwidget.id for fwidget in focused_widget.marked)
            if fids:
                log.debug('Found marked files: %r', fids)
                return fids

        # Get focused file IDs (if directory is focused, include all files in it)
        if hasattr(focused_widget, 'focused_file_ids'):
            fids = tuple(focused_widget.focused_file_ids)
            if fids:
                log.debug('Found focused files: %r', fids)
                return fids

        return ()


class create_list_widget():
    def create_list_widget(self, list_cls, *args, theme_name, markable_items=False, **kwargs):
        # Helper function that creates a tab title widget
        make_titlew = functools.partial(
            make_tab_title_widget,
            attr_unfocused='tabs.%s.unfocused' % theme_name,
            attr_focused='tabs.%s.focused' % theme_name)

        # If tab is specified by the user, pass it on to the list widget
        if hasattr(self, 'title'):
            kwargs['title'] = self.title

        # Create list widget
        log.debug('Creating %s(%s, %s)', list_cls.__name__, args, kwargs)
        from ...tui.tuiobjects import keymap
        listw = list_cls(objects.srvapi, keymap, *args, **kwargs)

        # Add list to tabs
        from ...tui.tuiobjects import tabs
        tabid = tabs.load(make_titlew(listw.title), listw)
        tabs.set_info(command=self.command)

        # If list items are markable, automatically add the 'marked' column
        if markable_items:
            columns = listw.columns
            if 'marked' not in columns:
                listw.columns = ('marked',) + columns

        # Tell list widget how to change the tab title
        def set_tab_title(text, count):
            # There is a race condition that can result in the list widget
            # setting a new title for a tab that has already been removed, which
            # raises an IndexError here.
            try:
                tabs.set_title(make_titlew(text, count), position=tabid)
            except IndexError:
                pass
        listw.title_updater = set_tab_title

        # Set a temporary title until the tab has finished loading its content
        # and the title_updater is called with the actual title
        tabs.set_title(make_titlew('Loading...', ''), position=tabid)


# TODO: Take a callable that return True to stop and False to keep polling rapidly.
#       `duration` should be a timeout to prevent polling rapidly forever.
class polling_frenzy():
    @classmethod
    def polling_frenzy(cls, duration=2, short_interval=0.5):
        """Reduce polling interval to `short_interval` for `duration` seconds"""
        if objects.srvapi.interval > 1:
            async def coro():
                log.debug('Setting poll interval to %s for %s seconds', short_interval, duration)
                orig_interval = objects.srvapi.interval
                objects.srvapi.interval = short_interval
                await asyncio.sleep(duration)
                objects.srvapi.interval = orig_interval
                log.debug('Interval restored to %s', objects.srvapi.interval)
            asyncio.ensure_future(coro())


class placeholders(make_request):
    class _Placeholder():
        def __init__(self, name, needed_keys, description, modifier=None):
            self.name = name
            self.needed_keys = needed_keys
            self.description = description
            self.modifier = modifier

        def evaluate(self, data):
            if self.modifier:
                return str(self.modifier(data))
            else:
                # assert len(self.needed_keys) == 1
                return str(data[self.needed_keys[0]])

    _placeholder_specs = {
        # {location}/{name} is absolute path to torrent file/directory
        'torrent': (_Placeholder('id', needed_keys=('id',),
                                 description='Torrent ID'),
                    _Placeholder('name', needed_keys=('name', 'comment'),
                                 description='Torrent name'),
                    _Placeholder('location', needed_keys=('path',),
                                 description='Download path (absolute)')),

        # {location}/{torrentname}/{path} is absolute path to file/directory
        # inside a torrent
        'file': (_Placeholder('id', needed_keys=('tid',),
                              description='Torrent ID'),
                 _Placeholder('name', needed_keys=('name',),
                              description='File or directory name without path'),
                 _Placeholder('location', needed_keys=('location',),
                              description='Torrent\'s absolute download path'),
                 _Placeholder('torrentname', needed_keys=('path-relative',),
                              description='Name of the torrent',
                              modifier=lambda data: data['path-relative'].split('/', maxsplit=1)[0]),
                 _Placeholder('path', needed_keys=('path-relative',),
                              description='Relative path within the torrent, including name',
                              modifier=lambda data: os.sep.join(data['path-relative'].split('/')[1:])))
    }

    HELP = ((('Placeholders are evaluated to values of the focused list item.  '
              'Their format is "{{NAME}}" where valid values for NAME are as follows:'),
             '',
             '\tIn torrent lists:')
            + tuple('\t\t%s \t- %s' % (ph.name, ph.description) for ph in _placeholder_specs['torrent'])
            + ('', '\tIn file lists:')
            + tuple('\t\t%s \t- %s' % (ph.name, ph.description) for ph in _placeholder_specs['file'])
            + ('',
               'To get a literal "{{", escape it with "\\".  "}}" doesn\'t need '
               'to be escaped.  Note that you must escape the "\\" itself unless '
               'it is quoted because it escapes arbitrary characters.'))

    _NOT_SUPPORTED_ERROR = 'Placeholders are not supported in the current tab'
    _RESOLVE_ERROR = 'Unable to resolve placeholders: %s'

    import re
    _placeholder_split_regex = re.compile(r'(?<!\\)(\{.+?\})')

    async def parse_placeholders(self, *args):
        parsed_args = []
        for i,arg in enumerate(args):
            next_arg = []
            for part in self._placeholder_split_regex.split(arg):
                log.debug('part: %r', part)
                if not part:
                    continue
                elif part.endswith('}'):
                    if part.startswith('{'):
                        next_arg.append(await self._resolve_placeholder(part[1:-1]))
                    else:
                        next_arg.append(part)
                else:
                    next_arg.append(part)

                # Unescape escaped curly brackets
                next_arg[-1] = next_arg[-1].replace('\\{', '{').replace('\\}', '}')

            parsed_args.append(''.join(next_arg))
        return parsed_args

    async def _resolve_placeholder(self, key):
        placeholders = await self._get_placeholder_map()
        try:
            return placeholders[key]
        except KeyError:
            raise CmdError('Unknown placeholder: %r' % key)

    async def _get_placeholder_map(self):
        if not hasattr(self, '_placeholders'):
            from ...tui.tuiobjects import tabs
            from ...tui.views import FileListWidget, TorrentListWidget
            focused_list = tabs.focus
            if focused_list is None:
                raise CmdError(self._RESOLVE_ERROR % 'No tab opened')
            elif not isinstance(focused_list, (TorrentListWidget, FileListWidget)):
                raise CmdError(self._NOT_SUPPORTED_ERROR)
            elif focused_list.focused_widget is None:
                raise CmdError(self._RESOLVE_ERROR % 'Current tab is empty')
            else:
                focused_item = focused_list.focused_widget.base_widget
                torrent_id = focused_item.torrent_id

                if isinstance(focused_list, TorrentListWidget):
                    phspecs = self._placeholder_specs['torrent']
                    needed_keys = sum((phspec.needed_keys for phspec in phspecs), ())
                    if all(key in focused_item.data for key in needed_keys):
                        # Cached Torrent object has everything we need
                        data = focused_item.data
                    else:
                        # Fetch data we need for placeholders
                        data = await self._fetch_torrent_data(torrent_id, needed_keys)
                elif isinstance(focused_list, FileListWidget):
                    phspecs = self._placeholder_specs['file']
                    # We don't need to fetch data because the file list item
                    # already has everything
                    data = focused_item.data

                else:
                    raise CmdError(self._NOT_SUPPORTED_ERROR)

                self._placeholders = {phspec.name:phspec.evaluate(data)
                                      for phspec in phspecs}
                log.debug('Placeholders: %r', self._placeholders)
        return self._placeholders

    async def _fetch_torrent_data(self, torrent_id, keys):
        log.debug('Fetching fresh Torrent #%d with keys: %r', torrent_id, keys)
        # Request new torrent because we can't be sure the wanted key
        # exists in widget.data
        request = objects.srvapi.torrent.torrents((torrent_id,), keys=keys)
        response = await self.make_request(request, quiet=True)
        if not response.success:
            raise CmdError()
        else:
            return response.torrents[0]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering Torrents by their values"""

from ..base import TorrentBase
from ..utils import Bandwidth, BoolOrBandwidth, Status, convert
from .base import BoolFilterSpec, CmpFilterSpec, Filter, FilterChain, FilterSpecDict
from .utils import cmp_timestamp_or_timdelta, limit_rate_filter, timestamp_or_timedelta


def _desc(text):
    if text.startswith('...'):
        text = 'Match VALUE against ' + text[4:]
    return text


_STATUS_VERIFY    = Status.VERIFY
_STATUS_DOWNLOAD  = Status.DOWNLOAD
_STATUS_UPLOAD    = Status.UPLOAD
_STATUS_INIT      = Status.INIT
_STATUS_CONNECTED = Status.CONNECTED
_STATUS_ISOLATED  = Status.ISOLATED
_STATUS_QUEUED    = Status.QUEUED
_STATUS_SEED      = Status.SEED
_STATUS_IDLE      = Status.IDLE
_STATUS_STOPPED   = Status.STOPPED


class _SingleFilter(Filter):
    DEFAULT_FILTER = 'name'

    BOOLEAN_FILTERS = FilterSpecDict({
        'all'         : BoolFilterSpec(None,
                                       aliases=('*',),
                                       description='All torrents'),
        'complete'    : BoolFilterSpec(lambda t: t['%downloaded'] >= 100,
                                       needed_keys=('%downloaded',),
                                       aliases=('cmp',),
                                       description='Torrents with all wanted files downloaded'),
        'stopped'     : BoolFilterSpec(lambda t: _STATUS_STOPPED in t['status'],
                                       needed_keys=('status',),
                                       aliases=('stp', 'paused'),
                                       description='Torrents that are not allowed to up- or download'),
        'active'      : BoolFilterSpec(lambda t: _STATUS_CONNECTED in t['status'] or _STATUS_VERIFY in t['status'],
                                       needed_keys=('peers-connected', 'status'),
                                       aliases=('act',),
                                       description='Torrents connected to peers or being verified'),
        'uploading'   : BoolFilterSpec(lambda t: t['rate-up'] > 0,
                                       needed_keys=('rate-up',),
                                       aliases=('upg',),
                                       description='Torrents using upload bandwidth'),
        'downloading' : BoolFilterSpec(lambda t: t['rate-down'] > 0,
                                       needed_keys=('rate-down',),
                                       aliases=('dng',),
                                       description='Torrents using download bandwidth'),
        'leeching'    : BoolFilterSpec(lambda t: t['%downloaded'] < 100 and _STATUS_STOPPED not in t['status'],
                                       needed_keys=('%downloaded', 'status'),
                                       aliases=('lcg',),
                                       description='Unstopped torrents downloading or waiting for seeds'),
        'seeding'     : BoolFilterSpec(lambda t: t['%downloaded'] >= 100 and _STATUS_STOPPED not in t['status'],
                                       needed_keys=('%downloaded', 'status'),
                                       aliases=('sdg',),
                                       description='Unstopped torrents with all wanted files downloaded'),
        'verifying'   : BoolFilterSpec(lambda t: _STATUS_VERIFY in t['status'],
                                       needed_keys=('status',),
                                       aliases=('vfg',),
                                       description='Torrents being verified or queued for verification'),
        'idle'        : BoolFilterSpec(lambda t: (_STATUS_IDLE in t['status'] and
                                                  _STATUS_STOPPED not in t['status']),
                                       needed_keys=('status',),
                                       description="Unstopped torrents that don't do anything"),
        'isolated'    : BoolFilterSpec(lambda t: _STATUS_ISOLATED in t['status'],
                                       needed_keys=('status',),
                                       aliases=('isl',),
                                       description='Torrents that cannot discover new peers in any way'),
        'private'     : BoolFilterSpec(lambda t: t['private'],
                                       needed_keys=('private',),
                                       aliases=('prv',),
                                       description='Torrents connectable through trackers only (no DHT/PEX)'),
    })

    COMPARATIVE_FILTERS = FilterSpecDict({
        'id'              : CmpFilterSpec(value_getter=lambda t: t['id'],
                                          value_type=TorrentBase.TYPES['id'],
                                          needed_keys=('id',),
                                          description=_desc('... torrent ID')),

        'infohash'        : CmpFilterSpec(value_getter=lambda t: t['hash'],
                                          value_type=TorrentBase.TYPES['hash'],
                                          needed_keys=('hash',),
                                          aliases=('hash',),
                                          description=_desc('... torrent SHA1 hash')),

        'name'            : CmpFilterSpec(value_getter=lambda t: t['name'],
                                          value_type=TorrentBase.TYPES['name'],
                                          needed_keys=('name',),
                                          aliases=('n',),
                                          description=_desc('... name')),

        'comment'         : CmpFilterSpec(value_getter=lambda t: t['comment'],
                                          value_type=TorrentBase.TYPES['comment'],
                                          needed_keys=('comment',),
                                          aliases=('cmnt',),
                                          description=_desc('... comment')),

        'path'            : CmpFilterSpec(value_getter=lambda t: t['path'],
                                          value_type=TorrentBase.TYPES['path'],
                                          needed_keys=('path',),
                                          description=_desc('... absolute path to download directory')),

        'error'           : CmpFilterSpec(value_getter=lambda t: t['error'],
                                          value_type=TorrentBase.TYPES['error'],
                                          needed_keys=('error',),
                                          aliases=('err',),
                                          description=_desc('... error message')),

        'uploaded'        : CmpFilterSpec(value_getter=lambda t: t['size-uploaded'],
                                          value_type=TorrentBase.TYPES['size-uploaded'],
                                          needed_keys=('size-uploaded',),
                                          aliases=('up',),
                                          description=_desc('... number of uploaded bytes')),

        'downloaded'      : CmpFilterSpec(value_getter=lambda t: t['size-downloaded'],
                                          value_type=TorrentBase.TYPES['size-downloaded'],
                                          needed_keys=('size-downloaded',),
                                          aliases=('dn',),
                                          description=_desc('... number of downloaded bytes')),

        '%downloaded'     : CmpFilterSpec(value_getter=lambda t: t['%downloaded'],
                                          value_type=TorrentBase.TYPES['%downloaded'],
                                          needed_keys=('%downloaded',),
                                          aliases=('%dn',),
                                          description=_desc('... percentage of downloaded bytes')),

        'size'            : CmpFilterSpec(value_getter=lambda t: t['size-final'],
                                          value_type=TorrentBase.TYPES['size-final'],
                                          value_convert=convert.size,
                                          needed_keys=('size-final',),
                                          aliases=('sz',),
                                          description=_desc('... combined size of all wanted files')),

        'peers'           : CmpFilterSpec(value_getter=lambda t: t['peers-connected'],
                                          value_type=TorrentBase.TYPES['peers-connected'],
                                          needed_keys=('peers-connected',),
                                          aliases=('prs',),
                                          description=_desc('... number of connected peers')),

        'seeds'           : CmpFilterSpec(value_getter=lambda t: t['peers-seeding'],
                                          value_type=TorrentBase.TYPES['peers-seeding'],
                                          needed_keys=('peers-seeding',),
                                          aliases=('sds',),
                                          description=_desc('... largest number of seeds reported by any tracker')),

        'ratio'           : CmpFilterSpec(value_getter=lambda t: t['ratio'],
                                          value_type=TorrentBase.TYPES['ratio'],
                                          needed_keys=('ratio',),
                                          aliases=('rto',),
                                          description=_desc('... uploaded/downloaded ratio')),

        'rate-up'         : CmpFilterSpec(value_getter=lambda t: t['rate-up'],
                                          value_type=TorrentBase.TYPES['rate-up'],
                                          value_convert=Bandwidth,
                                          needed_keys=('rate-up',),
                                          aliases=('rup',),
                                          description=_desc('... upload rate')),

        'rate-down'       : CmpFilterSpec(value_getter=lambda t: t['rate-down'],
                                          value_type=TorrentBase.TYPES['rate-down'],
                                          value_convert=Bandwidth,
                                          needed_keys=('rate-down',),
                                          aliases=('rdn',),
                                          description=_desc('... download rate')),

        'limit-rate-up'   : CmpFilterSpec(value_getter=lambda t: t['limit-rate-up'],
                                          value_matcher=lambda t, op, v: limit_rate_filter(t['limit-rate-up'], op, v),
                                          value_type=BoolOrBandwidth,
                                          as_bool=lambda t: t['limit-rate-up'] < float('inf'),
                                          needed_keys=('limit-rate-up',),
                                          aliases=('lrup',),
                                          description=_desc('... upload rate limit')),

        'limit-rate-down' : CmpFilterSpec(value_getter=lambda t: t['limit-rate-down'],
                                          value_matcher=lambda t, op, v: limit_rate_filter(t['limit-rate-down'], op, v),
                                          value_type=BoolOrBandwidth,
                                          as_bool=lambda t: t['limit-rate-down'] < float('inf'),
                                          needed_keys=('limit-rate-down',),
                                          aliases=('lrdn',),
                                          description=_desc('... download rate limit')),

        'tracker'         : CmpFilterSpec(value_getter=lambda t: (tracker['url-announce'].domain for tracker in t['trackers']),
                                          value_matcher=lambda t, op, v: any(op(tracker['url-announce'].domain, v)
                                                                             for tracker in t['trackers']),
                                          value_type=str,
                                          needed_keys=('trackers',),
                                          aliases=('trk',),
                                          description=_desc('... domain of the announce URL of trackers')),
        'label'           : CmpFilterSpec(value_getter=lambda t: t['labels'],
                                          value_matcher=lambda t, op, v:
                                              any(op(lbl, v) for lbl in t['labels']),
                                          value_type=str,
                                          needed_keys=('labels',),
                                          aliases=('lbl',),
                                          description=_desc('... labels')),

        'eta'             : CmpFilterSpec(value_getter=lambda t: t['timespan-eta'],
                                          value_matcher=lambda t, op, v: cmp_timestamp_or_timdelta(t['timespan-eta'], op, v),
                                          value_type=TorrentBase.TYPES['timespan-eta'],
                                          value_convert=timestamp_or_timedelta,
                                          needed_keys=('timespan-eta',),
                                          description=_desc('... estimated time to finish downloading')),

        'created'         : CmpFilterSpec(value_getter=lambda t: t['time-created'],
                                          value_matcher=lambda t, op, v: cmp_timestamp_or_timdelta(t['time-created'], op, v),
                                          value_type=TorrentBase.TYPES['time-created'],
                                          value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                          needed_keys=('time-created',),
                                          aliases=('tcrt',),
                                          description=_desc('... torrent creation time')),

        'added'           : CmpFilterSpec(value_getter=lambda t: t['time-added'],
                                          value_matcher=lambda t, op, v: cmp_timestamp_or_timdelta(t['time-added'], op, v),
                                          value_type=TorrentBase.TYPES['time-added'],
                                          value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                          needed_keys=('time-added',),
                                          aliases=('tadd',),
                                          description=_desc('... time torrent was added')),

        'started'         : CmpFilterSpec(value_getter=lambda t: t['time-started'],
                                          value_matcher=lambda t, op, v: cmp_timestamp_or_timdelta(t['time-started'], op, v),
                                          value_type=TorrentBase.TYPES['time-started'],
                                          value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                          needed_keys=('time-started',),
                                          aliases=('tsta',),
                                          description=_desc('... last time torrent was started')),

        'activity'        : CmpFilterSpec(value_getter=lambda t: t['time-activity'],
                                          value_matcher=lambda t, op, v: cmp_timestamp_or_timdelta(t['time-activity'], op, v),
                                          value_type=TorrentBase.TYPES['time-activity'],
                                          value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                          needed_keys=('time-activity',),
                                          aliases=('tact',),
                                          description=_desc('... time torrent was active')),

        'completed'       : CmpFilterSpec(value_getter=lambda t: t['time-completed'],
                                          value_matcher=lambda t, op, v: cmp_timestamp_or_timdelta(t['time-completed'], op, v),
                                          value_type=TorrentBase.TYPES['time-completed'],
                                          value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
                                          needed_keys=('time-completed',),
                                          aliases=('tcmp',),
                                          description=_desc('... time all wanted files where/will be downloaded')),
    })


class TorrentFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = _SingleFilter



# Load __version__ variable without importing the whole stig module
import re

from setuptools import find_packages, setup

version_match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]",
                          open('stig/__init__.py').read(), re.M)
if version_match:
    __version__ = version_match.group(1)
else:
    raise RuntimeError("Unable to find __version__")

setup(
    name             = 'stig',
    version          = __version__,
    license          = 'GPLv3+',
    author           = 'Random User',
    author_email     = 'rndusr@posteo.de',

    description      = 'TUI and CLI client for the Transmission daemon',
    long_description = (
        "There is no `long_description` because PyPI doesn't support org-mode "
        "conversion with `pandoc` is wonky and I can't be arsed to convert "
        "the README by hand."
        "\n\n"
        "Please read the README in the [repo](https://github.com/rndusr/stig)."
    ),
    long_description_content_type='text/markdown',

    url              = 'https://github.com/rndusr/stig',
    keywords         = 'bittorrent torrent transmission',

    packages         = find_packages(),
    package_data     = {'stig': ['settings/default.theme']},

    python_requires  = '>=3.6',
    install_requires = [
        'urwid>=2.6.12',
        'urwidtrees==1.0.3',
        'aiohttp==3.*',
        'async_timeout',
        'pyxdg',
        'blinker',
        'natsort',
        'wcwidth',
    ],
    extras_require = {
        'setproctitle': ['setproctitle'],
        'proxy': ['aiohttp-socks'],
    },
    tests_require = [
        'pytest>=5,<6',
        'asynctest>=0.11',
    ],

    entry_points = { 'console_scripts': [ 'stig = stig:run' ] },

    classifiers = [
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Environment :: Console',
        'Operating System :: Unix',
        'Development Status :: 3 - Alpha',
    ],
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...objects import localcfg
from .file import TUICOLUMNS as FCOLUMNS
from .peer import TUICOLUMNS as PCOLUMNS
from .torrent import TUICOLUMNS as TCOLUMNS

_BANDWIDTH_COLUMNS = (TCOLUMNS['rate-up'], TCOLUMNS['rate-down'],
                      TCOLUMNS['limit-rate-up'], TCOLUMNS['limit-rate-down'],
                      PCOLUMNS['rate-up'], PCOLUMNS['rate-down'], PCOLUMNS['rate-est'])
def _set_bandwidth_unit(settings, name, value):
    unit_short = {'bit': 'b', 'byte': 'B'}.get(value, value)
    unit_short += '/s'
    for column in _BANDWIDTH_COLUMNS:
        column.set_header(right=unit_short)
localcfg.on_change(_set_bandwidth_unit, name='unit.bandwidth')


_SIZE_COLUMNS = (TCOLUMNS['size'], TCOLUMNS['downloaded'],
                 TCOLUMNS['uploaded'], TCOLUMNS['available'],
                 FCOLUMNS['size'], FCOLUMNS['downloaded'])
def _set_size_unit(settings, name, value):
    unit_short = {'bit': 'b', 'byte': 'B'}.get(value, value)
    for column in _SIZE_COLUMNS:
        column.set_header(right=unit_short)
localcfg.on_change(_set_size_unit, name='unit.size')
